# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :func:`iris_grib._save_rules._missing_forecast_period.`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris.cube import Cube
from iris.coords import DimCoord

from iris_grib._save_rules import _missing_forecast_period


class TestNoForecastReferenceTime(tests.IrisGribTest):
    def test_no_bounds(self):
        t_coord = DimCoord(15, "time", units="hours since epoch")
        cube = Cube(23)
        cube.add_aux_coord(t_coord)

        res = _missing_forecast_period(cube)
        expected_rt = t_coord.units.num2date(15)
        expected_rt_type = 3
        expected_fp = 0
        expected_fp_type = 1
        expected = (expected_rt, expected_rt_type, expected_fp, expected_fp_type)
        self.assertEqual(res, expected)

    def test_with_bounds(self):
        t_coord = DimCoord(15, "time", bounds=[14, 16], units="hours since epoch")
        cube = Cube(23)
        cube.add_aux_coord(t_coord)

        res = _missing_forecast_period(cube)
        expected_rt = t_coord.units.num2date(14)
        expected_rt_type = 3
        expected_fp = 0
        expected_fp_type = 1
        expected = (expected_rt, expected_rt_type, expected_fp, expected_fp_type)
        self.assertEqual(res, expected)


class TestWithForecastReferenceTime(tests.IrisGribTest):
    def test_no_bounds(self):
        t_coord = DimCoord(3, "time", units="days since epoch")
        frt_coord = DimCoord(8, "forecast_reference_time", units="hours since epoch")
        cube = Cube(23)
        cube.add_aux_coord(t_coord)
        cube.add_aux_coord(frt_coord)

        res = _missing_forecast_period(cube)
        expected_rt = frt_coord.units.num2date(8)
        expected_rt_type = 1
        expected_fp = 3 * 24 - 8
        expected_fp_type = 1
        expected = (expected_rt, expected_rt_type, expected_fp, expected_fp_type)
        self.assertEqual(res, expected)

    def test_with_bounds(self):
        t_coord = DimCoord(3, "time", bounds=[2, 4], units="days since epoch")
        frt_coord = DimCoord(8, "forecast_reference_time", units="hours since epoch")
        cube = Cube(23)
        cube.add_aux_coord(t_coord)
        cube.add_aux_coord(frt_coord)

        res = _missing_forecast_period(cube)
        expected_rt = frt_coord.units.num2date(8)
        expected_rt_type = 1
        expected_fp = 2 * 24 - 8
        expected_fp_type = 1
        expected = (expected_rt, expected_rt_type, expected_fp, expected_fp_type)
        self.assertEqual(res, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.forecast_period_coord.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris.coords import DimCoord

from iris_grib._load_convert import forecast_period_coord


class Test(tests.IrisGribTest):
    def test(self):
        # (indicatorOfUnitForForecastTime, forecastTime, expected-hours)
        times = [
            (0, 60, 1),  # minutes
            (1, 2, 2),  # hours
            (2, 1, 24),  # days
            (10, 2, 6),  # 3 hours
            (11, 3, 18),  # 6 hours
            (12, 2, 24),  # 12 hours
            (13, 3600, 1),
        ]  # seconds

        for indicatorOfUnitForForecastTime, forecastTime, hours in times:
            coord = forecast_period_coord(indicatorOfUnitForForecastTime, forecastTime)
            self.assertIsInstance(coord, DimCoord)
            self.assertEqual(coord.standard_name, "forecast_period")
            self.assertEqual(coord.units, "hours")
            self.assertEqual(coord.shape, (1,))
            self.assertEqual(coord.points[0], hours)
            self.assertFalse(coord.has_bounds())


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_40`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_40


class _Section(dict):
    def get_computed_key(self, key):
        return self.get(key)


class Test_regular(tests.IrisGribTest):
    def section_3(self):
        section = _Section(
            {
                "shapeOfTheEarth": 0,
                "scaleFactorOfRadiusOfSphericalEarth": 0,
                "scaledValueOfRadiusOfSphericalEarth": 6367470,
                "scaleFactorOfEarthMajorAxis": 0,
                "scaledValueOfEarthMajorAxis": MDI,
                "scaleFactorOfEarthMinorAxis": 0,
                "scaledValueOfEarthMinorAxis": MDI,
                "iDirectionIncrement": 22500000,
                "longitudeOfFirstGridPoint": 0,
                "resolutionAndComponentFlags": 32,
                "Ni": 16,
                "scanningMode": 0b01000000,
                "distinctLatitudes": np.array(
                    [
                        -73.79921363,
                        -52.81294319,
                        -31.70409175,
                        -10.56988231,
                        10.56988231,
                        31.70409175,
                        52.81294319,
                        73.79921363,
                    ]
                ),
                "numberOfOctectsForNumberOfPoints": 0,
                "interpretationOfNumberOfPoints": 0,
            }
        )
        return section

    def expected(self, y_dim, x_dim, y_neg=True):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        nx = 16
        dx = 22.5
        x_origin = 0
        x = iris.coords.DimCoord(
            np.arange(nx) * dx + x_origin,
            standard_name="longitude",
            units="degrees_east",
            coord_system=cs,
            circular=True,
        )
        y_points = np.array(
            [
                73.79921363,
                52.81294319,
                31.70409175,
                10.56988231,
                -10.56988231,
                -31.70409175,
                -52.81294319,
                -73.79921363,
            ]
        )
        if not y_neg:
            y_points = y_points[::-1]
        y = iris.coords.DimCoord(
            y_points, standard_name="latitude", units="degrees_north", coord_system=cs
        )
        expected["dim_coords_and_dims"].append((y, y_dim))
        expected["dim_coords_and_dims"].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_40(section, metadata)
        expected = self.expected(0, 1, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_transposed(self):
        section = self.section_3()
        section["scanningMode"] = 0b01100000
        metadata = empty_metadata()
        grid_definition_template_40(section, metadata)
        expected = self.expected(1, 0, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_reverse_latitude(self):
        section = self.section_3()
        section["scanningMode"] = 0b00000000
        metadata = empty_metadata()
        grid_definition_template_40(section, metadata)
        expected = self.expected(0, 1, y_neg=True)
        self.assertEqual(metadata, expected)


class Test_reduced(tests.IrisGribTest):
    def section_3(self):
        section = _Section(
            {
                "shapeOfTheEarth": 0,
                "scaleFactorOfRadiusOfSphericalEarth": 0,
                "scaledValueOfRadiusOfSphericalEarth": 6367470,
                "scaleFactorOfEarthMajorAxis": 0,
                "scaledValueOfEarthMajorAxis": MDI,
                "scaleFactorOfEarthMinorAxis": 0,
                "scaledValueOfEarthMinorAxis": MDI,
                "longitudes": np.array(
                    [0.0, 180.0, 0.0, 120.0, 240.0, 0.0, 120.0, 240.0, 0.0, 180.0]
                ),
                "latitudes": np.array(
                    [
                        -59.44440829,
                        -59.44440829,
                        -19.87571915,
                        -19.87571915,
                        -19.87571915,
                        19.87571915,
                        19.87571915,
                        19.87571915,
                        59.44440829,
                        59.44440829,
                    ]
                ),
                "numberOfOctectsForNumberOfPoints": 1,
                "interpretationOfNumberOfPoints": 1,
            }
        )
        return section

    def expected(self):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        x_points = np.array(
            [0.0, 180.0, 0.0, 120.0, 240.0, 0.0, 120.0, 240.0, 0.0, 180.0]
        )
        y_points = np.array(
            [
                -59.44440829,
                -59.44440829,
                -19.87571915,
                -19.87571915,
                -19.87571915,
                19.87571915,
                19.87571915,
                19.87571915,
                59.44440829,
                59.44440829,
            ]
        )
        x = iris.coords.AuxCoord(
            x_points, standard_name="longitude", units="degrees_east", coord_system=cs
        )
        y = iris.coords.AuxCoord(
            y_points, standard_name="latitude", units="degrees_north", coord_system=cs
        )
        expected["aux_coords_and_dims"].append((y, 0))
        expected["aux_coords_and_dims"].append((x, 0))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        expected = self.expected()
        grid_definition_template_40(section, metadata)
        self.assertEqual(metadata, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function
:func:`iris_grib._load_convert.grid_definition_template_0_and_1`.

"""

# Import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords

from iris.exceptions import TranslationError

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import grid_definition_template_0_and_1


class _Section(dict):
    def get_computed_key(self, key):
        return self.get(key)


class Test_resolution_flags(tests.IrisGribTest):
    def section_3(self):
        section = _Section(
            {
                "Ni": 6,
                "Nj": 6,
                "latitudeOfFirstGridPoint": 0,
                "longitudeOfFirstGridPoint": 0,
                "resolutionAndComponentFlags": 0,
                "latitudeOfLastGridPoint": 5000000,
                "longitudeOfLastGridPoint": 5000000,
                "iDirectionIncrement": 0,
                "jDirectionIncrement": 0,
                "scanningMode": 0b01000000,
                "numberOfOctectsForNumberOfPoints": 0,
                "interpretationOfNumberOfPoints": 0,
            }
        )
        return section

    def expected(self, x_dim, y_dim, x_points, y_points, x_neg=True, y_neg=True):
        # Prepare the expectation.
        expected = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        if x_neg:
            x_points = x_points[::-1]
        x = iris.coords.DimCoord(
            x_points, standard_name="longitude", units="degrees", coord_system=cs
        )
        if y_neg:
            y_points = y_points[::-1]
        y = iris.coords.DimCoord(
            y_points, standard_name="latitude", units="degrees", coord_system=cs
        )
        expected["dim_coords_and_dims"].append((y, y_dim))
        expected["dim_coords_and_dims"].append((x, x_dim))
        return expected

    def test_without_increments(self):
        section = self.section_3()
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, "latitude", "longitude", cs)
        x_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        y_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_with_increments(self):
        section = self.section_3()
        section["resolutionAndComponentFlags"] = 48
        section["iDirectionIncrement"] = 1000000
        section["jDirectionIncrement"] = 1000000
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, "latitude", "longitude", cs)
        x_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        y_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_with_i_not_j_increment(self):
        section = self.section_3()
        section["resolutionAndComponentFlags"] = 32
        section["iDirectionIncrement"] = 1000000
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, "latitude", "longitude", cs)
        x_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        y_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_with_j_not_i_increment(self):
        section = self.section_3()
        section["resolutionAndComponentFlags"] = 16
        section["jDirectionIncrement"] = 1000000
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, "latitude", "longitude", cs)
        x_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        y_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False, y_neg=False)
        self.assertEqual(metadata, expected)

    def test_without_increments_crossing_0_lon(self):
        section = self.section_3()
        section["longitudeOfFirstGridPoint"] = 355000000
        section["Ni"] = 11
        metadata = empty_metadata()
        cs = iris.coord_systems.GeogCS(6367470)
        grid_definition_template_0_and_1(section, metadata, "latitude", "longitude", cs)
        x_points = np.array(
            [
                355.0,
                356.0,
                357.0,
                358.0,
                359.0,
                360.0,
                361.0,
                362.0,
                363.0,
                364.0,
                365.0,
            ]
        )
        y_points = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
        expected = self.expected(1, 0, x_points, y_points, x_neg=False, y_neg=False)
        self.assertEqual(metadata, expected)


class Test(tests.IrisGribTest):
    def test_unsupported_quasi_regular__number_of_octets(self):
        section = {"numberOfOctectsForNumberOfPoints": 1}
        cs = None
        metadata = None
        with self.assertRaisesRegex(TranslationError, "quasi-regular"):
            grid_definition_template_0_and_1(
                section, metadata, "latitude", "longitude", cs
            )

    def test_unsupported_quasi_regular__interpretation(self):
        section = {
            "numberOfOctectsForNumberOfPoints": 1,
            "interpretationOfNumberOfPoints": 1,
        }
        cs = None
        metadata = None
        with self.assertRaisesRegex(TranslationError, "quasi-regular"):
            grid_definition_template_0_and_1(
                section, metadata, "latitude", "longitude", cs
            )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.set_time_range`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock
import warnings

from cf_units import Unit
import eccodes

from iris.coords import DimCoord

from iris_grib._save_rules import set_time_range


class Test(tests.IrisGribTest):
    def setUp(self):
        self.coord = DimCoord(
            0, "time", units=Unit("hours since epoch", calendar="standard")
        )

    def test_no_bounds(self):
        with self.assertRaisesRegex(
            ValueError, "Expected time coordinate " "with two bounds, got 0 bounds"
        ):
            set_time_range(self.coord, mock.sentinel.grib)

    def test_three_bounds(self):
        self.coord.bounds = [0, 1, 2]
        with self.assertRaisesRegex(
            ValueError, "Expected time coordinate " "with two bounds, got 3 bounds"
        ):
            set_time_range(self.coord, mock.sentinel.grib)

    def test_non_scalar(self):
        coord = DimCoord(
            [0, 1],
            "time",
            bounds=[[0, 1], [1, 2]],
            units=Unit("hours since epoch", calendar="standard"),
        )
        with self.assertRaisesRegex(
            ValueError, "Expected length one time " "coordinate, got 2 points"
        ):
            set_time_range(coord, mock.sentinel.grib)

    @mock.patch.object(eccodes, "codes_set")
    def test_hours(self, mock_set):
        lower = 10
        upper = 20
        self.coord.bounds = [lower, upper]
        set_time_range(self.coord, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib, "indicatorOfUnitForTimeRange", 1)
        mock_set.assert_any_call(mock.sentinel.grib, "lengthOfTimeRange", upper - lower)

    @mock.patch.object(eccodes, "codes_set")
    def test_days(self, mock_set):
        lower = 4
        upper = 6
        self.coord.bounds = [lower, upper]
        self.coord.units = Unit("days since epoch", calendar="standard")
        set_time_range(self.coord, mock.sentinel.grib)
        mock_set.assert_any_call(mock.sentinel.grib, "indicatorOfUnitForTimeRange", 1)
        mock_set.assert_any_call(
            mock.sentinel.grib, "lengthOfTimeRange", (upper - lower) * 24
        )

    @mock.patch.object(eccodes, "codes_set")
    def test_fractional_hours(self, mock_set_long):
        lower = 10.0
        upper = 20.9
        self.coord.bounds = [lower, upper]
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            set_time_range(self.coord, mock.sentinel.grib)
        self.assertEqual(len(warn), 1)
        msg = (
            r"Truncating floating point lengthOfTimeRange 10\.8?9+ "
            "to integer value 10"
        )
        self.assertRegex(str(warn[0].message), msg)
        mock_set_long.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeRange", 1
        )
        mock_set_long.assert_any_call(
            mock.sentinel.grib, "lengthOfTimeRange", int(upper - lower)
        )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""Unit tests for `iris_grib.grib_save_rules.reference_time`."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import eccodes

from iris_grib import load_cubes
from iris_grib._save_rules import reference_time


class Test(tests.IrisGribTest):
    def _test(self, cube):
        grib = mock.Mock()
        mock_eccodes = mock.Mock(spec=eccodes)
        with mock.patch("iris_grib._save_rules.eccodes", mock_eccodes):
            reference_time(cube, grib)

        mock_eccodes.assert_has_calls(
            [
                mock.call.codes_set_long(grib, "significanceOfReferenceTime", 1),
                mock.call.codes_set_long(grib, "dataDate", "19980306"),
                mock.call.codes_set_long(grib, "dataTime", "0300"),
            ]
        )

    @tests.skip_data
    def test_forecast_period(self):
        # The stock cube has a non-compliant forecast_period.
        fname = tests.get_data_path(("GRIB", "global_t", "global.grib2"))
        [cube] = load_cubes(fname)
        self._test(cube)

    @tests.skip_data
    def test_no_forecast_period(self):
        # The stock cube has a non-compliant forecast_period.
        fname = tests.get_data_path(("GRIB", "global_t", "global.grib2"))
        [cube] = load_cubes(fname)
        cube.remove_coord("forecast_period")
        self._test(cube)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""Unit tests for the :mod:`iris_grib._save_rules` package."""



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for function :func:`iris_grib._load_convert.statistical_forecast_period`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import datetime
from unittest import mock

from iris_grib._load_convert import statistical_forecast_period_coord


class Test(tests.IrisGribTest):
    def setUp(self):
        module = "iris_grib._load_convert"
        self.module = module
        self.patch_hindcast = self.patch(module + "._hindcast_fix")
        self.forecast_seconds = 0.0
        self.forecast_units = mock.Mock()
        self.forecast_units.convert = lambda x, y: self.forecast_seconds
        self.patch(module + ".time_range_unit", return_value=self.forecast_units)
        self.frt_coord = mock.Mock()
        self.frt_coord.points = [1]
        self.frt_coord.units.num2date = mock.Mock(
            return_value=datetime.datetime(2010, 2, 3)
        )
        self.section = {}
        self.section["yearOfEndOfOverallTimeInterval"] = 2010
        self.section["monthOfEndOfOverallTimeInterval"] = 2
        self.section["dayOfEndOfOverallTimeInterval"] = 3
        self.section["hourOfEndOfOverallTimeInterval"] = 8
        self.section["minuteOfEndOfOverallTimeInterval"] = 0
        self.section["secondOfEndOfOverallTimeInterval"] = 0
        self.section["forecastTime"] = mock.Mock()
        self.section["indicatorOfUnitForForecastTime"] = mock.Mock()

    def test_basic(self):
        coord = statistical_forecast_period_coord(self.section, self.frt_coord)
        self.assertEqual(coord.standard_name, "forecast_period")
        self.assertEqual(coord.units, "hours")
        self.assertArrayAlmostEqual(coord.points, [4.0])
        self.assertArrayAlmostEqual(coord.bounds, [[0.0, 8.0]])

    def test_with_hindcast(self):
        _ = statistical_forecast_period_coord(self.section, self.frt_coord)
        self.assertEqual(self.patch_hindcast.call_count, 1)

    def test_no_hindcast(self):
        self.patch(self.module + ".options.support_hindcast_values", False)
        _ = statistical_forecast_period_coord(self.section, self.frt_coord)
        self.assertEqual(self.patch_hindcast.call_count, 0)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for function :func:`iris_grib._load_convert.translate_phenomenon`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from cf_units import Unit
from iris.coords import DimCoord

from iris_grib._load_convert import Probability, translate_phenomenon
from iris_grib.grib_phenom_translation import Grib1CfData, GRIBCode


class Test_probability(tests.IrisGribTest):
    def setUp(self):
        # Patch inner call to return a given phenomenon type.
        target_module = "iris_grib._load_convert"
        self.phenom_lookup_patch = self.patch(
            target_module + ".itranslation.grib2_phenom_to_cf_info",
            return_value=Grib1CfData("air_temperature", "", "K", None),
        )
        # Construct dummy call arguments
        self.probability = Probability("<prob_type>", 22.0)
        self.metadata = {"aux_coords_and_dims": [], "attributes": {}}

    def test_basic(self):
        translate_phenomenon(
            self.metadata, 7, 8, 9, None, None, None, probability=self.probability
        )
        # Check metadata.
        thresh_coord = DimCoord(
            [22.0], standard_name="air_temperature", long_name="", units="K"
        )
        self.assertEqual(
            self.metadata,
            {
                "standard_name": None,
                "long_name": "probability_of_air_temperature_<prob_type>",
                "units": Unit(1),
                "aux_coords_and_dims": [(thresh_coord, None)],
                "attributes": {"GRIB_PARAM": GRIBCode(2, 7, 8, 9)},
            },
        )

    def test_no_phenomenon(self):
        self.phenom_lookup_patch.return_value = None
        expected_metadata = self.metadata.copy()
        translate_phenomenon(
            self.metadata,
            discipline=7,
            parameterCategory=77,
            parameterNumber=777,
            typeOfFirstFixedSurface=None,
            scaledValueOfFirstFixedSurface=None,
            typeOfSecondFixedSurface=None,
            probability=self.probability,
        )
        expected_metadata["attributes"]["GRIB_PARAM"] = GRIBCode(2, 7, 77, 777)
        self.assertEqual(self.metadata, expected_metadata)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_10`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coord_systems
import iris.coords
import iris.exceptions

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import grid_definition_template_10


class Test(tests.IrisGribTest):
    def section_3(self):
        section = {
            "gridDefinitionTemplateNumber": 10,
            "shapeOfTheEarth": 1,
            "scaleFactorOfRadiusOfSphericalEarth": 0,
            "scaledValueOfRadiusOfSphericalEarth": 6371200,
            "scaleFactorOfEarthMajorAxis": 0,
            "scaledValueOfEarthMajorAxis": 0,
            "scaleFactorOfEarthMinorAxis": 0,
            "scaledValueOfEarthMinorAxis": 0,
            "Ni": 181,
            "Nj": 213,
            "latitudeOfFirstGridPoint": 2351555,
            "latitudeOfLastGridPoint": 25088204,
            "LaD": 14000000,
            "longitudeOfFirstGridPoint": 114990304,
            "longitudeOfLastGridPoint": 135009712,
            "resolutionAndComponentFlags": 56,
            "scanningMode": 64,
            "Di": 12000000,
            "Dj": 12000000,
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(6371200.0)
        cs = iris.coord_systems.Mercator(standard_parallel=14.0, ellipsoid=ellipsoid)
        nx = 181
        x_origin = 12406918.990644248
        dx = 12000
        x = iris.coords.DimCoord(
            np.arange(nx) * dx + x_origin,
            "projection_x_coordinate",
            units="m",
            coord_system=cs,
        )
        ny = 213
        y_origin = 253793.10903714459

        dy = 12000
        y = iris.coords.DimCoord(
            np.arange(ny) * dy + y_origin,
            "projection_y_coordinate",
            units="m",
            coord_system=cs,
        )
        expected["dim_coords_and_dims"].append((y, y_dim))
        expected["dim_coords_and_dims"].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_10(section, metadata)
        expected = self.expected(y_dim=0, x_dim=1)
        self.assertEqual(metadata, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function
:func:`iris_grib._load_convert.product_definition_template_0`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

import iris.coords

import iris_grib
from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib.tests.unit.load_convert import LoadConvertTest
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import product_definition_template_0


def section_4():
    return {
        "hoursAfterDataCutoff": MDI,
        "minutesAfterDataCutoff": MDI,
        "indicatorOfUnitForForecastTime": 0,  # minutes
        "forecastTime": 360,
        "NV": 0,
        "typeOfFirstFixedSurface": 103,
        "scaleFactorOfFirstFixedSurface": 0,
        "scaledValueOfFirstFixedSurface": 9999,
        "typeOfSecondFixedSurface": 255,
    }


class Test(LoadConvertTest):
    def test_given_frt(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(
            24, "forecast_reference_time", units="hours since epoch"
        )
        product_definition_template_0(section_4(), metadata, rt_coord)
        expected = empty_metadata()
        aux = expected["aux_coords_and_dims"]
        aux.append((iris.coords.DimCoord(6, "forecast_period", units="hours"), None))
        aux.append((iris.coords.DimCoord(30, "time", units="hours since epoch"), None))
        aux.append((rt_coord, None))
        aux.append((iris.coords.DimCoord(9999, long_name="height", units="m"), None))
        self.assertMetadataEqual(metadata, expected)

    def test_given_t(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(24, "time", units="hours since epoch")
        product_definition_template_0(section_4(), metadata, rt_coord)
        expected = empty_metadata()
        aux = expected["aux_coords_and_dims"]
        aux.append((iris.coords.DimCoord(6, "forecast_period", units="hours"), None))
        aux.append(
            (
                iris.coords.DimCoord(
                    18, "forecast_reference_time", units="hours since epoch"
                ),
                None,
            )
        )
        aux.append((rt_coord, None))
        aux.append((iris.coords.DimCoord(9999, long_name="height", units="m"), None))
        self.assertMetadataEqual(metadata, expected)

    def test_generating_process_warnings(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(
            24, "forecast_reference_time", units="hours since epoch"
        )
        convert_options = iris_grib._load_convert.options
        emit_warnings = convert_options.warn_on_unsupported
        try:
            convert_options.warn_on_unsupported = True
            with mock.patch("warnings.warn") as warn:
                product_definition_template_0(section_4(), metadata, rt_coord)
            warn_msgs = [call[1][0] for call in warn.mock_calls]
            expected = [
                "Unable to translate type of generating process.",
                "Unable to translate background generating process " "identifier.",
                "Unable to translate forecast generating process " "identifier.",
            ]
            self.assertEqual(warn_msgs, expected)
        finally:
            convert_options.warn_on_unsupported = emit_warnings


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_140`.
"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import cartopy.crs as ccrs
import numpy as np

import iris.coord_systems
import iris.coords

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_140


class Test(tests.IrisGribTest):
    def section_3(self):
        section = {
            "gridDefinitionTemplateNumber": 140,
            "shapeOfTheEarth": 4,
            "scaleFactorOfRadiusOfSphericalEarth": MDI,
            "scaledValueOfRadiusOfSphericalEarth": MDI,
            "scaleFactorOfEarthMajorAxis": MDI,
            "scaledValueOfEarthMajorAxis": MDI,
            "scaleFactorOfEarthMinorAxis": MDI,
            "scaledValueOfEarthMinorAxis": MDI,
            "numberOfPointsAlongXAxis": 2,
            "numberOfPointsAlongYAxis": 2,
            "latitudeOfFirstGridPoint": 53988880,
            "longitudeOfFirstGridPoint": -4027984,
            "standardParallelInMicrodegrees": 54900000,
            "centralLongitudeInMicrodegrees": -2500000,
            "resolutionAndComponentFlags": 0b00110000,
            "xDirectionGridLengthInMillimetres": 2000000,
            "yDirectionGridLengthInMillimetres": 2000000,
            "scanningMode": 0b01000000,
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(6378137, inverse_flattening=298.257222101)
        cs = iris.coord_systems.LambertAzimuthalEqualArea(
            latitude_of_projection_origin=54.9,
            longitude_of_projection_origin=-2.5,
            false_easting=0,
            false_northing=0,
            ellipsoid=ellipsoid,
        )
        lon0 = -4027984 * 1e-6
        lat0 = 53988880 * 1e-6
        x0m, y0m = cs.as_cartopy_crs().transform_point(lon0, lat0, ccrs.Geodetic())
        dxm = dym = 2000.0
        x_points = x0m + dxm * np.arange(2)
        y_points = y0m + dym * np.arange(2)
        x = iris.coords.DimCoord(
            x_points,
            standard_name="projection_x_coordinate",
            units="m",
            coord_system=cs,
        )
        y = iris.coords.DimCoord(
            y_points,
            standard_name="projection_y_coordinate",
            units="m",
            coord_system=cs,
        )
        expected["dim_coords_and_dims"].append((y, y_dim))
        expected["dim_coords_and_dims"].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_140(section, metadata)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""Unit tests for the `iris_grib.save_messages` function."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import eccodes
from unittest import mock

import iris_grib


class TestSaveMessages(tests.IrisGribTest):
    def setUp(self):
        # Create a test object to stand in for a real PPField.
        self.grib_message = eccodes.codes_grib_new_from_samples("GRIB2")

    def test_save(self):
        m = mock.mock_open()
        with mock.patch("builtins.open", m, create=True):
            iris_grib.save_messages([self.grib_message], "foo.grib2")
        self.assertTrue(mock.call("foo.grib2", "wb") in m.mock_calls)

    def test_save_append(self):
        m = mock.mock_open()
        with mock.patch("builtins.open", m, create=True):
            iris_grib.save_messages([self.grib_message], "foo.grib2", append=True)
        self.assertTrue(mock.call("foo.grib2", "ab") in m.mock_calls)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Integration test for round-trip loading and saving of various grids.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from typing import Literal

import numpy as np

from iris import load_cube, save
from iris.coord_systems import GeogCS, PolarStereographic, RotatedGeogCS, Stereographic
from iris.coords import DimCoord
from iris.cube import Cube
from iris.fileformats.pp import EARTH_RADIUS as UM_DEFAULT_EARTH_RADIUS
from iris.util import is_regular

from iris_grib.grib_phenom_translation import GRIBCode


class TestGDT5(tests.TestGribMessage):
    def test_save_load(self):
        # Load sample UKV data (variable-resolution rotated grid).
        path = tests.get_data_path(("PP", "ukV1", "ukVpmslont.pp"))
        cube = load_cube(path)

        # Extract a single 2D field, for simplicity.
        self.assertEqual(cube.ndim, 3)
        self.assertEqual(cube.coord_dims("time"), (0,))
        cube = cube[0]

        # Check that it has a rotated-pole variable-spaced grid, as expected.
        x_coord = cube.coord(axis="x")
        self.assertIsInstance(x_coord.coord_system, RotatedGeogCS)
        self.assertFalse(is_regular(x_coord))

        # Write to temporary file, check that key contents are in the file,
        # then load back in.
        with self.temp_filename("ukv_sample.grib2") as temp_file_path:
            save(cube, temp_file_path)

            # Check that various aspects of the saved file are as expected.
            expect_values = (
                (0, "editionNumber", 2),
                (3, "gridDefinitionTemplateNumber", 5),
                (3, "Ni", cube.shape[-1]),
                (3, "Nj", cube.shape[-2]),
                (3, "shapeOfTheEarth", 1),
                (
                    3,
                    "scaledValueOfRadiusOfSphericalEarth",
                    int(UM_DEFAULT_EARTH_RADIUS),
                ),
                (3, "resolutionAndComponentFlags", 0),
                (3, "latitudeOfSouthernPole", -37500000),
                (3, "longitudeOfSouthernPole", 357500000),
                (3, "angleOfRotation", 0),
            )
            self.assertGribMessageContents(temp_file_path, expect_values)

            # Load the Grib file back into a new cube.
            cube_loaded_from_saved = load_cube(temp_file_path)
            # Also load data, before the temporary file gets deleted.
            cube_loaded_from_saved.data

        # The re-loaded result will not match the original in every respect:
        #  * cube attributes are discarded
        #  * horizontal coordinates are rounded to an integer representation
        #  * bounds on horizontal coords are lost
        # Thus the following "equivalence tests" are rather piecemeal..

        # Check those re-loaded properties which should match the original.
        for test_cube in (cube, cube_loaded_from_saved):
            self.assertEqual(test_cube.standard_name, "air_pressure_at_sea_level")
            self.assertEqual(test_cube.units, "Pa")
            self.assertEqual(test_cube.shape, (928, 744))
            self.assertEqual(test_cube.cell_methods, ())

        # Check only the GRIB_PARAM attribute exists on the re-loaded cube.
        # Note: this does *not* match the original, but is as expected.
        self.assertEqual(
            cube_loaded_from_saved.attributes,
            {"GRIB_PARAM": GRIBCode("GRIB2:d000c003n001")},
        )

        # Now remaining to check: coordinates + data...

        # Check they have all the same coordinates.
        co_names = [coord.name() for coord in cube.coords()]
        co_names_reload = [coord.name() for coord in cube_loaded_from_saved.coords()]
        self.assertEqual(sorted(co_names_reload), sorted(co_names))

        # Check all the coordinates.
        for coord_name in co_names:
            try:
                co_orig = cube.coord(coord_name)
                co_load = cube_loaded_from_saved.coord(coord_name)

                # Check shape.
                self.assertEqual(
                    co_load.shape,
                    co_orig.shape,
                    'Shape of re-loaded "{}" coord is {} ' "instead of {}".format(
                        coord_name, co_load.shape, co_orig.shape
                    ),
                )

                # Check coordinate points equal, within a tolerance.
                self.assertArrayAllClose(co_load.points, co_orig.points, rtol=1.0e-6)

                # Check all coords are unbounded.
                # (NOTE: this is not so for the original X and Y coordinates,
                # but Grib does not store those bounds).
                self.assertIsNone(co_load.bounds)

            except AssertionError as err:
                self.assertTrue(
                    False,
                    'Failed on coordinate "{}" : {}'.format(coord_name, str(err)),
                )

        # Check that main data array also matches.
        self.assertArrayAllClose(cube.data, cube_loaded_from_saved.data)


class GDT20Common:
    """'Wrapper' to avoid class being detected during test runs."""

    class GDT20Common(tests.TestGribMessage):
        """Roundtrip testing that attributes save+load correctly."""

        # Enable subclassing to test different permutations.
        coord_system_class: type[Stereographic]
        pole: Literal["north", "south"]

        def setUp(self):
            # Create a Cube to save, inspired by
            #  iris-sample-data toa_brightness_stereographic.nc
            if self.pole == "north":
                central_lat = 90
            elif self.pole == "south":
                central_lat = -90
            else:
                raise NotImplementedError(f"Invalid pole: {self.pole}")

            self.coord_system_kwargs = dict(
                central_lat=central_lat,
                central_lon=325,
                true_scale_lat=central_lat,
                ellipsoid=GeogCS(6378169.0),
            )
            coord_system = self.coord_system_class(**self.coord_system_kwargs)

            coord_kwargs = dict(
                units="m",
                coord_system=coord_system,
            )
            coord_x = DimCoord(
                np.linspace(-2250000, 6750192, 256, endpoint=False),
                standard_name="projection_x_coordinate",
                **coord_kwargs,
            )
            coord_y = DimCoord(
                np.linspace(-980000, -6600000, 160, endpoint=False),
                standard_name="projection_y_coordinate",
                **coord_kwargs,
            )
            coord_t = DimCoord(
                0, standard_name="time", units="hours since 1970-01-01 00:00:00"
            )
            coord_fp = DimCoord(0, standard_name="forecast_period", units="hours")
            coord_frt = DimCoord(
                0, standard_name="forecast_reference_time", units=coord_t.units
            )
            shape = (coord_y.shape[0], coord_x.shape[0])
            self.cube = Cube(
                np.arange(np.prod(shape), dtype=float).reshape(shape),
                dim_coords_and_dims=[(coord_y, 0), (coord_x, 1)],
                aux_coords_and_dims=[
                    (coord_t, None),
                    (coord_fp, None),
                    (coord_frt, None),
                ],
            )

        def test_save_load(self):
            with self.temp_filename("polar_stereo.grib2") as temp_file_path:
                save(self.cube, temp_file_path)
                cube_reloaded = load_cube(temp_file_path)
                # Touch the data before destroying the file.
                _ = cube_reloaded.data

            cube_expected = self.cube.copy()
            for coord in cube_expected.dim_coords:
                # GRIB only describes PolarStereographic, so we always expect
                #  that system even when we started with Stereographic.
                coord.coord_system = PolarStereographic(**self.coord_system_kwargs)

            # Modifications to remove irrelevant inequalities.
            del cube_reloaded.attributes["GRIB_PARAM"]
            for coord in cube_reloaded.dim_coords:
                coord.points = np.round(coord.points)

            self.assertEqual(cube_expected, cube_reloaded)


class TestGDT20StereoNorth(GDT20Common.GDT20Common):
    coord_system_class = Stereographic
    pole = "north"


class TestGDT20StereoSouth(GDT20Common.GDT20Common):
    coord_system_class = Stereographic
    pole = "south"


class TestGDT20PolarNorth(GDT20Common.GDT20Common):
    coord_system_class = PolarStereographic
    pole = "north"


class TestGDT20PolarSouth(GDT20Common.GDT20Common):
    coord_system_class = PolarStereographic
    pole = "south"


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Integration tests to confirm data is loaded correctly.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np
import numpy.ma as ma

from iris import load_cube


class TestImport(tests.IrisGribTest):
    def test_gdt1(self):
        path = tests.get_data_path(("GRIB", "rotated_nae_t", "sensible_pole.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)

    def test_gdt90_with_bitmap(self):
        path = tests.get_data_path(("GRIB", "umukv", "ukv_chan9.grib2"))
        cube = load_cube(path)
        # Pay particular attention to the orientation.
        self.assertIsNot(cube.data[0, 0], ma.masked)
        self.assertIs(cube.data[-1, 0], ma.masked)
        self.assertIs(cube.data[0, -1], ma.masked)
        self.assertIs(cube.data[-1, -1], ma.masked)
        x = cube.coord("projection_x_coordinate").points
        y = cube.coord("projection_y_coordinate").points
        self.assertGreater(x[0], x[-1])  # Decreasing X coordinate
        self.assertLess(y[0], y[-1])  # Increasing Y coordinate
        # Check everything else.
        self.assertCMLApproxData(cube)


class TestGDT30(tests.IrisGribTest):
    def test_lambert(self):
        path = tests.get_data_path(("GRIB", "lambert", "lambert.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)


class TestGDT40(tests.IrisGribTest):
    def test_regular(self):
        path = tests.get_data_path(("GRIB", "gaussian", "regular_gg.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)

    def test_reduced(self):
        path = tests.get_data_path(("GRIB", "reduced", "reduced_gg.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)


class TestDRT3(tests.IrisGribTest):
    def test_grid_complex_spatial_differencing(self):
        path = tests.get_data_path(("GRIB", "missing_values", "missing_values.grib2"))
        cube = load_cube(path)
        self.assertCMLApproxData(cube)


class TestDataProxy(tests.IrisGribTest):
    def test_data_representation__no_bitsPerValue(self):
        """Ensures that zero data is auto-generated."""
        # This test file contains one GRIB2 message with no data payload
        # in Data Section [7], but has "bitsPerValue=0" in the Data
        # Representation Section [5] to trigger auto-generation of zero
        # data payload by the DataProxy.
        path = tests.get_data_path(
            ("GRIB", "missing_values", "ice_severity__no_bitsPerValue.grib2")
        )
        cube = load_cube(path)
        self.assertCML(cube)
        self.assertEqual(0, np.sum(cube.data))


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.ellipsoid_geometry.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from iris_grib._load_convert import ellipsoid_geometry


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section = {
            "scaledValueOfEarthMajorAxis": 10,
            "scaleFactorOfEarthMajorAxis": 1,
            "scaledValueOfEarthMinorAxis": 100,
            "scaleFactorOfEarthMinorAxis": 2,
            "scaledValueOfRadiusOfSphericalEarth": 1000,
            "scaleFactorOfRadiusOfSphericalEarth": 3,
        }

    def test_geometry(self):
        result = ellipsoid_geometry(self.section)
        self.assertEqual(result, (1.0, 1.0, 1.0))


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :meth:`iris_grib._save_rules.grid_definition_template_140`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coords
from iris.coord_systems import GeogCS, LambertAzimuthalEqualArea
from iris.exceptions import TranslationError

from iris_grib._save_rules import (
    grid_definition_template_140 as grid_definition_template,
)
from iris_grib.tests.unit.save_rules import GdtTestMixin


class FakeGribError(Exception):
    pass


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        self.default_ellipsoid = GeogCS(
            semi_major_axis=6377563.396, semi_minor_axis=6356256.909
        )
        self.test_cube = self._make_test_cube()

        GdtTestMixin.setUp(self)

    def _make_test_cube(self, cs=None, x_points=None, y_points=None):
        # Create a cube with given properties, or minimal defaults.
        if cs is None:
            cs = self._default_coord_system()
        if x_points is None:
            x_points = self._default_x_points()
        if y_points is None:
            y_points = self._default_y_points()

        x_coord = iris.coords.DimCoord(
            x_points, "projection_x_coordinate", units="m", coord_system=cs
        )
        y_coord = iris.coords.DimCoord(
            y_points, "projection_y_coordinate", units="m", coord_system=cs
        )
        test_cube = iris.cube.Cube(np.zeros((len(y_points), len(x_points))))
        test_cube.add_dim_coord(y_coord, 0)
        test_cube.add_dim_coord(x_coord, 1)
        return test_cube

    def _default_coord_system(self, false_easting=0, false_northing=0):
        return LambertAzimuthalEqualArea(
            latitude_of_projection_origin=54.9,
            longitude_of_projection_origin=-2.5,
            false_easting=false_easting,
            false_northing=false_northing,
            ellipsoid=self.default_ellipsoid,
        )

    def test__template_number(self):
        grid_definition_template(self.test_cube, self.mock_grib)
        self._check_key("gridDefinitionTemplateNumber", 140)

    def test__shape_of_earth(self):
        grid_definition_template(self.test_cube, self.mock_grib)
        self._check_key("shapeOfTheEarth", 7)
        self._check_key("scaleFactorOfEarthMajorAxis", 0)
        self._check_key("scaledValueOfEarthMajorAxis", 6377563.396)
        self._check_key("scaleFactorOfEarthMinorAxis", 0)
        self._check_key("scaledValueOfEarthMinorAxis", 6356256.909)

    def test__grid_shape(self):
        test_cube = self._make_test_cube(x_points=np.arange(13), y_points=np.arange(6))
        grid_definition_template(test_cube, self.mock_grib)
        self._check_key("Nx", 13)
        self._check_key("Ny", 6)

    def test__grid_points(self):
        test_cube = self._make_test_cube(
            x_points=[1e6, 3e6, 5e6, 7e6], y_points=[4e6, 9e6]
        )
        grid_definition_template(test_cube, self.mock_grib)
        self._check_key("latitudeOfFirstGridPoint", 81330008)
        self._check_key("longitudeOfFirstGridPoint", 98799008)
        self._check_key("Dx", 2e9)
        self._check_key("Dy", 5e9)

    # specific to grid_definition_template_140
    def test__template_specifics(self):
        grid_definition_template(self.test_cube, self.mock_grib)
        self._check_key("standardParallelInMicrodegrees", 54900000)
        self._check_key("centralLongitudeInMicrodegrees", 357500000)

    def test__scanmode(self):
        grid_definition_template(self.test_cube, self.mock_grib)
        self._check_scanmode(+1, +1)

    def test__scanmode_reverse(self):
        test_cube = self._make_test_cube(x_points=np.arange(7e6, 0, -1e6))
        grid_definition_template(test_cube, self.mock_grib)
        self._check_scanmode(-1, +1)

    def __fail_false_easting_northing(self, false_easting, false_northing):
        cs = self._default_coord_system(
            false_easting=false_easting, false_northing=false_northing
        )
        test_cube = self._make_test_cube(cs=cs)
        msg = (
            r"non zero false easting \(\d*\.\d{2}\) or "
            r"non zero false northing \(\d*\.\d{2}\)"
            r"; unsupported by GRIB Template 3\.140"
            r""
        )
        with self.assertRaisesRegex(TranslationError, msg):
            grid_definition_template(test_cube, self.mock_grib)

    def test__fail_false_easting(self):
        self.__fail_false_easting_northing(10.0, 0.0)

    def test__fail_false_northing(self):
        self.__fail_false_easting_northing(0.0, 10.0)

    def test__fail_false_easting_northing(self):
        self.__fail_false_easting_northing(10.0, 10.0)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_20`.
"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import cartopy.crs as ccrs
import numpy as np

import iris.coord_systems
import iris.coords

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_20


class Test(tests.IrisGribTest):
    def section_3(self):
        section = {
            "gridDefinitionTemplateNumber": 20,
            "shapeOfTheEarth": 0,
            "scaleFactorOfRadiusOfSphericalEarth": 0,
            "scaledValueOfRadiusOfSphericalEarth": 6367470,
            "scaleFactorOfEarthMajorAxis": 0,
            "scaledValueOfEarthMajorAxis": MDI,
            "scaleFactorOfEarthMinorAxis": 0,
            "scaledValueOfEarthMinorAxis": MDI,
            "Nx": 15,
            "Ny": 10,
            "latitudeOfFirstGridPoint": 32549114,
            "longitudeOfFirstGridPoint": 225385728,
            "resolutionAndComponentFlags": 0b00001000,
            "LaD": 60000000,
            "orientationOfTheGrid": 262000000,
            "Dx": 320000000,
            "Dy": 320000000,
            "projectionCentreFlag": 0b00000000,
            "scanningMode": 0b01000000,
        }
        return section

    def expected(self, y_dim, x_dim):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(6367470)
        # Always expect PolarStereographic - never Stereographic.
        #  Stereographic is a CF/Iris concept and not something described in
        #  GRIB.
        cs = iris.coord_systems.PolarStereographic(
            central_lat=90.0,
            central_lon=262.0,
            false_easting=0,
            false_northing=0,
            true_scale_lat=60.0,
            ellipsoid=ellipsoid,
        )
        lon0 = 225385728 * 1e-6
        lat0 = 32549114 * 1e-6
        x0m, y0m = cs.as_cartopy_crs().transform_point(lon0, lat0, ccrs.Geodetic())
        dxm = dym = 320000.0
        x_points = x0m + dxm * np.arange(15)
        y_points = y0m + dym * np.arange(10)
        x = iris.coords.DimCoord(
            x_points,
            standard_name="projection_x_coordinate",
            units="m",
            coord_system=cs,
            circular=False,
        )
        y = iris.coords.DimCoord(
            y_points,
            standard_name="projection_y_coordinate",
            units="m",
            coord_system=cs,
        )
        expected["dim_coords_and_dims"].append((y, y_dim))
        expected["dim_coords_and_dims"].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_20(section, metadata)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""Unit tests for :func:`iris_grib._grib1_load_rules.grib1_convert`."""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else
import iris_grib.tests as tests

import eccodes
from unittest import mock

from iris.aux_factory import HybridPressureFactory
from iris.exceptions import TranslationError
from iris.fileformats.rules import Reference

from iris_grib import GribWrapper
from iris_grib._grib1_load_rules import grib1_convert
from iris_grib.tests.unit import TestField


class TestBadEdition(tests.IrisGribTest):
    def test(self):
        message = mock.Mock(edition=2)
        emsg = "GRIB edition 2 is not supported"
        with self.assertRaisesRegex(TranslationError, emsg):
            grib1_convert(message)


class TestBoundedTime(TestField):
    @staticmethod
    def is_forecast_period(coord):
        return coord.standard_name == "forecast_period" and coord.units == "hours"

    @staticmethod
    def is_time(coord):
        return coord.standard_name == "time" and coord.units == "hours since epoch"

    def assert_bounded_message(self, **kwargs):
        attributes = {
            "productDefinitionTemplateNumber": 0,
            "edition": 1,
            "_forecastTime": 15,
            "_forecastTimeUnit": "hours",
            "phenomenon_bounds": lambda u: (80, 120),
            "_phenomenonDateTime": -1,
            "table2Version": 9999,
            "_originatingCentre": "xxx",
        }
        attributes.update(kwargs)
        message = mock.Mock(**attributes)
        self._test_for_coord(
            message,
            grib1_convert,
            self.is_forecast_period,
            expected_points=[35],
            expected_bounds=[[15, 55]],
        )
        self._test_for_coord(
            message,
            grib1_convert,
            self.is_time,
            expected_points=[100],
            expected_bounds=[[80, 120]],
        )

    def assert_bounded_message_3hours(self, **kwargs):
        attributes = {
            "productDefinitionTemplateNumber": 0,
            "edition": 1,
            "_forecastTime": 252,
            "_forecastTimeUnit": "3 hours",
            "phenomenon_bounds": lambda u: (252, 258),
            "_phenomenonDateTime": -1,
            "table2Version": 9999,
            "_originatingCentre": "xxx",
        }
        attributes.update(kwargs)
        message = mock.Mock(**attributes)
        self._test_for_coord(
            message,
            grib1_convert,
            self.is_forecast_period,
            expected_points=[255],
            expected_bounds=[[252, 258]],
        )
        self._test_for_coord(
            message,
            grib1_convert,
            self.is_time,
            expected_points=[255],
            expected_bounds=[[252, 258]],
        )

    def test_time_range_indicator_2(self):
        self.assert_bounded_message(timeRangeIndicator=2)
        self.assert_bounded_message_3hours(timeRangeIndicator=2)

    def test_time_range_indicator_3(self):
        self.assert_bounded_message(timeRangeIndicator=3)
        self.assert_bounded_message_3hours(timeRangeIndicator=3)

    def test_time_range_indicator_4(self):
        self.assert_bounded_message(timeRangeIndicator=4)
        self.assert_bounded_message_3hours(timeRangeIndicator=4)

    def test_time_range_indicator_5(self):
        self.assert_bounded_message(timeRangeIndicator=5)
        self.assert_bounded_message_3hours(timeRangeIndicator=5)

    def test_time_range_indicator_51(self):
        self.assert_bounded_message(timeRangeIndicator=51)
        self.assert_bounded_message_3hours(timeRangeIndicator=51)

    def test_time_range_indicator_113(self):
        self.assert_bounded_message(timeRangeIndicator=113)
        self.assert_bounded_message_3hours(timeRangeIndicator=113)

    def test_time_range_indicator_114(self):
        self.assert_bounded_message(timeRangeIndicator=114)
        self.assert_bounded_message_3hours(timeRangeIndicator=114)

    def test_time_range_indicator_115(self):
        self.assert_bounded_message(timeRangeIndicator=115)
        self.assert_bounded_message_3hours(timeRangeIndicator=115)

    def test_time_range_indicator_116(self):
        self.assert_bounded_message(timeRangeIndicator=116)
        self.assert_bounded_message_3hours(timeRangeIndicator=116)

    def test_time_range_indicator_117(self):
        self.assert_bounded_message(timeRangeIndicator=117)
        self.assert_bounded_message_3hours(timeRangeIndicator=117)

    def test_time_range_indicator_118(self):
        self.assert_bounded_message(timeRangeIndicator=118)
        self.assert_bounded_message_3hours(timeRangeIndicator=118)

    def test_time_range_indicator_123(self):
        self.assert_bounded_message(timeRangeIndicator=123)
        self.assert_bounded_message_3hours(timeRangeIndicator=123)

    def test_time_range_indicator_124(self):
        self.assert_bounded_message(timeRangeIndicator=124)
        self.assert_bounded_message_3hours(timeRangeIndicator=124)

    def test_time_range_indicator_125(self):
        self.assert_bounded_message(timeRangeIndicator=125)
        self.assert_bounded_message_3hours(timeRangeIndicator=125)


class Test_GribLevels(tests.IrisTest):
    def test_grib1_hybrid_height(self):
        gm = eccodes.codes_grib_new_from_samples("regular_gg_ml_grib1")
        gw = GribWrapper(gm)
        results = grib1_convert(gw)

        (factory,) = results[0]
        self.assertEqual(factory.factory_class, HybridPressureFactory)
        delta, sigma, ref = factory.args
        self.assertEqual(delta, {"long_name": "level_pressure"})
        self.assertEqual(sigma, {"long_name": "sigma"})
        self.assertEqual(ref, Reference(name="surface_pressure"))

        coords_and_dims = results[8]
        (coord,) = [
            co for co, _ in coords_and_dims if co.name() == "model_level_number"
        ]
        self.assertEqual(coord.units, "1")
        self.assertEqual(coord.attributes["positive"], "up")
        (coord,) = [co for co, _ in coords_and_dims if co.name() == "level_pressure"]
        self.assertEqual(coord.units, "Pa")
        (coord,) = [co for co, _ in coords_and_dims if co.name() == "sigma"]
        self.assertEqual(coord.units, "1")


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for the `iris_grib.GribWrapper` class.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np
from unittest import mock

from iris._lazy_data import is_lazy_data
from iris.exceptions import TranslationError

from iris_grib import GribWrapper, _load_generate


_offset = 0
_message_length = 1000


def _mock_codes_get_long(grib_message, key):
    lookup = dict(
        totalLength=_message_length,
        numberOfValues=200,
        jPointsAreConsecutive=0,
        Ni=20,
        Nj=10,
        edition=1,
    )
    try:
        result = lookup[key]
    except KeyError:
        msg = "Mock codes_get_long unknown key: {!r}".format(key)
        raise AttributeError(msg)
    return result


def _mock_codes_get_string(grib_message, key):
    return grib_message


def _mock_codes_get_native_type(grib_message, key):
    result = int
    if key == "gridType":
        result = str
    return result


def _mock_codes_get_message_offset(grib_message):
    return _offset


class Test_edition(tests.IrisGribTest):
    def setUp(self):
        self.patch("iris_grib.GribWrapper._confirm_in_scope")
        self.patch("iris_grib.GribWrapper._compute_extra_keys")
        self.patch("eccodes.codes_get_long", _mock_codes_get_long)
        self.patch("eccodes.codes_get_string", _mock_codes_get_string)
        self.patch("eccodes.codes_get_native_type", _mock_codes_get_native_type)
        self.patch("eccodes.codes_get_message_offset", _mock_codes_get_message_offset)

    def test_not_edition_1(self):
        def func(grib_message, key):
            return 2

        emsg = "GRIB edition 2 is not supported by 'GribWrapper'"
        with mock.patch("eccodes.codes_get_long", func):
            with self.assertRaisesRegex(TranslationError, emsg):
                GribWrapper(None)

    def test_edition_1(self):
        grib_message = "regular_ll"
        grib_fh = mock.Mock()
        wrapper = GribWrapper(grib_message, grib_fh)
        self.assertEqual(wrapper.grib_message, grib_message)


@tests.skip_data
class Test_deferred_data(tests.IrisTest):
    def test_regular_data(self):
        filename = tests.get_data_path(("GRIB", "gaussian", "regular_gg.grib1"))
        messages = list(_load_generate(filename))
        self.assertTrue(is_lazy_data(messages[0]._data))

    def test_reduced_data(self):
        filename = tests.get_data_path(("GRIB", "reduced", "reduced_ll.grib1"))
        messages = list(_load_generate(filename))
        self.assertTrue(is_lazy_data(messages[0]._data))


class Test_deferred_proxy_args(tests.IrisTest):
    def setUp(self):
        self.patch("iris_grib.GribWrapper._confirm_in_scope")
        self.patch("iris_grib.GribWrapper._compute_extra_keys")
        self.patch("eccodes.codes_get_long", _mock_codes_get_long)
        self.patch("eccodes.codes_get_string", _mock_codes_get_string)
        self.patch("eccodes.codes_get_native_type", _mock_codes_get_native_type)
        self.patch("eccodes.codes_get_message_offset", _mock_codes_get_message_offset)
        self.expected = np.atleast_1d(_offset)
        self.grib_fh = mock.Mock()
        self.dtype = np.float64
        self.path = self.grib_fh.name
        self.lookup = _mock_codes_get_long

    def test_regular_proxy_args(self):
        grib_message = "regular_ll"
        shape = (self.lookup(grib_message, "Nj"), self.lookup(grib_message, "Ni"))
        for offset in self.expected:
            with mock.patch("iris_grib.GribDataProxy") as mock_gdp:
                _ = GribWrapper(grib_message, self.grib_fh)
            mock_gdp.assert_called_once_with(shape, self.dtype, self.path, offset)

    def test_reduced_proxy_args(self):
        grib_message = "reduced_gg"
        shape = self.lookup(grib_message, "numberOfValues")
        for offset in self.expected:
            with mock.patch("iris_grib.GribDataProxy") as mock_gdp:
                _ = GribWrapper(grib_message, self.grib_fh)
            mock_gdp.assert_called_once_with((shape,), self.dtype, self.path, offset)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.

"""
Provides GRIB/CF phenomenon translations.

"""

from collections import namedtuple


CFName = namedtuple("CFName", "standard_name long_name units")

DimensionCoordinate = namedtuple("DimensionCoordinate", "standard_name units points")

G1LocalParam = namedtuple("G1LocalParam", "edition t2version centre iParam")
G2Param = namedtuple("G2Param", "edition discipline category number")


GRIB1_LOCAL_TO_CF_CONSTRAINED = {
    G1LocalParam(1, 128, 98, 165): (
        CFName("x_wind", None, "m s-1"),
        DimensionCoordinate("height", "m", (10,)),
    ),
    G1LocalParam(1, 128, 98, 166): (
        CFName("y_wind", None, "m s-1"),
        DimensionCoordinate("height", "m", (10,)),
    ),
    G1LocalParam(1, 128, 98, 167): (
        CFName("air_temperature", None, "K"),
        DimensionCoordinate("height", "m", (2,)),
    ),
    G1LocalParam(1, 128, 98, 168): (
        CFName("dew_point_temperature", None, "K"),
        DimensionCoordinate("height", "m", (2,)),
    ),
}

GRIB1_LOCAL_TO_CF = {
    G1LocalParam(1, 128, 98, 31): CFName("sea_ice_area_fraction", None, "1"),
    G1LocalParam(1, 128, 98, 34): CFName("sea_surface_temperature", None, "K"),
    G1LocalParam(1, 128, 98, 59): CFName(
        "atmosphere_specific_convective_available_potential_energy", None, "J kg-1"
    ),
    G1LocalParam(1, 128, 98, 129): CFName("geopotential", None, "m2 s-2"),
    G1LocalParam(1, 128, 98, 130): CFName("air_temperature", None, "K"),
    G1LocalParam(1, 128, 98, 131): CFName("x_wind", None, "m s-1"),
    G1LocalParam(1, 128, 98, 132): CFName("y_wind", None, "m s-1"),
    G1LocalParam(1, 128, 98, 135): CFName(
        "lagrangian_tendency_of_air_pressure", None, "Pa s-1"
    ),
    G1LocalParam(1, 128, 98, 141): CFName("thickness_of_snowfall_amount", None, "m"),
    G1LocalParam(1, 128, 98, 151): CFName("air_pressure_at_sea_level", None, "Pa"),
    G1LocalParam(1, 128, 98, 157): CFName("relative_humidity", None, "%"),
    G1LocalParam(1, 128, 98, 164): CFName("cloud_area_fraction", None, "1"),
    G1LocalParam(1, 128, 98, 173): CFName("surface_roughness_length", None, "m"),
    G1LocalParam(1, 128, 98, 174): CFName(None, "grib_physical_atmosphere_albedo", "1"),
    G1LocalParam(1, 128, 98, 186): CFName("low_type_cloud_area_fraction", None, "1"),
    G1LocalParam(1, 128, 98, 187): CFName("medium_type_cloud_area_fraction", None, "1"),
    G1LocalParam(1, 128, 98, 188): CFName("high_type_cloud_area_fraction", None, "1"),
    G1LocalParam(1, 128, 98, 235): CFName(None, "grib_skin_temperature", "K"),
}

GRIB2_TO_CF = {
    G2Param(2, 0, 0, 0): CFName("air_temperature", None, "K"),
    G2Param(2, 0, 0, 2): CFName("air_potential_temperature", None, "K"),
    G2Param(2, 0, 0, 6): CFName("dew_point_temperature", None, "K"),
    G2Param(2, 0, 0, 10): CFName("surface_upward_latent_heat_flux", None, "W m-2"),
    G2Param(2, 0, 0, 11): CFName("surface_upward_sensible_heat_flux", None, "W m-2"),
    G2Param(2, 0, 0, 17): CFName("surface_temperature", None, "K"),
    G2Param(2, 0, 0, 32): CFName("wet_bulb_potential_temperature", None, "K"),
    G2Param(2, 0, 1, 0): CFName("specific_humidity", None, "kg kg-1"),
    G2Param(2, 0, 1, 1): CFName("relative_humidity", None, "%"),
    G2Param(2, 0, 1, 2): CFName("humidity_mixing_ratio", None, "kg kg-1"),
    G2Param(2, 0, 1, 3): CFName(None, "precipitable_water", "kg m-2"),
    G2Param(2, 0, 1, 7): CFName("precipitation_flux", None, "kg m-2 s-1"),
    G2Param(2, 0, 1, 9): CFName(
        "stratiform_rainfall_amount",
        "Large-scale precipitation (non-convective)",
        "kg m-2",
    ),
    G2Param(2, 0, 1, 10): CFName(
        "convective_rainfall_amount", "Convective precipitation", "kg m-2"
    ),
    G2Param(2, 0, 1, 11): CFName("thickness_of_snowfall_amount", None, "m"),
    G2Param(2, 0, 1, 13): CFName(
        "liquid_water_content_of_surface_snow", None, "kg m-2"
    ),
    G2Param(2, 0, 1, 15): CFName(
        "stratiform_snowfall_amount", "Large-scale snow", "kg m-2"
    ),
    G2Param(2, 0, 1, 22): CFName(None, "cloud_mixing_ratio", "kg kg-1"),
    G2Param(2, 0, 1, 37): CFName(
        "convective_rainfall_flux", "Convective precipitation rate", "kg m-2 s-1"
    ),
    G2Param(2, 0, 1, 49): CFName("precipitation_amount", None, "kg m-2"),
    G2Param(2, 0, 1, 51): CFName("atmosphere_mass_content_of_water", None, "kg m-2"),
    G2Param(2, 0, 1, 53): CFName("snowfall_flux", None, "kg m-2 s-1"),
    G2Param(2, 0, 1, 58): CFName(
        "convective_snowfall_flux", "Convective snowfall rate", "kg m-2 s-1"
    ),
    G2Param(2, 0, 1, 59): CFName(
        "stratiform_snowfall_flux", "Large scale snowfall rate", "kg m-2 s-1"
    ),
    G2Param(2, 0, 1, 60): CFName("snowfall_amount", None, "kg m-2"),
    G2Param(2, 0, 1, 64): CFName(
        "atmosphere_mass_content_of_water_vapor", None, "kg m-2"
    ),
    G2Param(2, 0, 1, 77): CFName(
        "stratiform_rainfall_flux", "Large scale rain rate", "kg m-2 s-1"
    ),
    G2Param(2, 0, 1, 83): CFName(
        "mass_fraction_of_cloud_liquid_water_in_air", None, "kg kg-1"
    ),
    G2Param(2, 0, 1, 84): CFName("mass_fraction_of_cloud_ice_in_air", None, "kg kg-1"),
    G2Param(2, 0, 2, 0): CFName("wind_from_direction", None, "degrees"),
    G2Param(2, 0, 2, 1): CFName("wind_speed", None, "m s-1"),
    G2Param(2, 0, 2, 2): CFName("x_wind", None, "m s-1"),
    G2Param(2, 0, 2, 3): CFName("y_wind", None, "m s-1"),
    G2Param(2, 0, 2, 8): CFName("lagrangian_tendency_of_air_pressure", None, "Pa s-1"),
    G2Param(2, 0, 2, 9): CFName(
        "upward_air_velocity", "Vertical velocity (geometric)", "m s-1"
    ),
    G2Param(2, 0, 2, 10): CFName("atmosphere_absolute_vorticity", None, "s-1"),
    G2Param(2, 0, 2, 12): CFName(
        "atmosphere_relative_vorticity", "Relative vorticity", "s-1"
    ),
    G2Param(2, 0, 2, 14): CFName(None, "ertel_potential_velocity", "K m2 kg-1 s-1"),
    G2Param(2, 0, 2, 22): CFName("wind_speed_of_gust", None, "m s-1"),
    G2Param(2, 0, 3, 0): CFName("air_pressure", None, "Pa"),
    G2Param(2, 0, 3, 1): CFName("air_pressure_at_sea_level", None, "Pa"),
    G2Param(2, 0, 3, 3): CFName(None, "icao_standard_atmosphere_reference_height", "m"),
    G2Param(2, 0, 3, 4): CFName("geopotential", None, "m2 s-2"),
    G2Param(2, 0, 3, 5): CFName("geopotential_height", None, "m"),
    G2Param(2, 0, 3, 6): CFName("altitude", None, "m"),
    G2Param(2, 0, 3, 9): CFName("geopotential_height_anomaly", None, "m"),
    G2Param(2, 0, 4, 7): CFName(
        "surface_downwelling_shortwave_flux_in_air", None, "W m-2"
    ),
    G2Param(2, 0, 4, 9): CFName("surface_net_downward_shortwave_flux", None, "W m-2"),
    G2Param(2, 0, 5, 3): CFName(
        "surface_downwelling_longwave_flux_in_air", None, "W m-2"
    ),
    G2Param(2, 0, 5, 4): CFName(
        "toa_outgoing_longwave_flux", "Upward long-wave radiation flux", "W m-2"
    ),
    G2Param(2, 0, 5, 5): CFName("surface_net_downward_longwave_flux", None, "W m-2"),
    G2Param(2, 0, 6, 1): CFName(
        None, "cloud_area_fraction_assuming_maximum_random_overlap", "1"
    ),
    G2Param(2, 0, 6, 3): CFName("low_type_cloud_area_fraction", None, "%"),
    G2Param(2, 0, 6, 4): CFName("medium_type_cloud_area_fraction", None, "%"),
    G2Param(2, 0, 6, 5): CFName("high_type_cloud_area_fraction", None, "%"),
    G2Param(2, 0, 6, 6): CFName(
        "atmosphere_mass_content_of_cloud_liquid_water", None, "kg m-2"
    ),
    G2Param(2, 0, 6, 7): CFName("cloud_area_fraction_in_atmosphere_layer", None, "%"),
    G2Param(2, 0, 6, 25): CFName(None, "WAFC_CB_horizontal_extent", "1"),
    G2Param(2, 0, 6, 26): CFName(None, "WAFC_ICAO_height_at_cloud_base", "m"),
    G2Param(2, 0, 6, 27): CFName(None, "WAFC_ICAO_height_at_cloud_top", "m"),
    G2Param(2, 0, 7, 6): CFName(
        "atmosphere_specific_convective_available_potential_energy", None, "J kg-1"
    ),
    G2Param(2, 0, 7, 7): CFName(None, "convective_inhibition", "J kg-1"),
    G2Param(2, 0, 7, 8): CFName(None, "storm_relative_helicity", "J kg-1"),
    G2Param(2, 0, 14, 0): CFName("atmosphere_mole_content_of_ozone", None, "Dobson"),
    G2Param(2, 0, 19, 1): CFName(None, "grib_physical_atmosphere_albedo", "%"),
    G2Param(2, 0, 19, 20): CFName(None, "WAFC_icing_potential", "1"),
    G2Param(2, 0, 19, 21): CFName(None, "WAFC_in-cloud_turb_potential", "1"),
    G2Param(2, 0, 19, 22): CFName(None, "WAFC_CAT_potential", "1"),
    G2Param(2, 2, 0, 0): CFName("land_binary_mask", None, "1"),
    G2Param(2, 2, 0, 0): CFName("land_area_fraction", None, "1"),
    G2Param(2, 2, 0, 1): CFName("surface_roughness_length", None, "m"),
    G2Param(2, 2, 0, 2): CFName("soil_temperature", None, "K"),
    G2Param(2, 2, 0, 3): CFName(
        "soil_moisture_content", "Soil moisture content", "kg m-2"
    ),
    G2Param(2, 2, 0, 7): CFName("surface_altitude", None, "m"),
    G2Param(2, 2, 0, 22): CFName("moisture_content_of_soil_layer", None, "kg m-2"),
    G2Param(2, 2, 0, 34): CFName("surface_runoff_flux", None, "kg m-2 s-1"),
    G2Param(2, 10, 1, 2): CFName("sea_water_x_velocity", None, "m s-1"),
    G2Param(2, 10, 1, 3): CFName("sea_water_y_velocity", None, "m s-1"),
    G2Param(2, 10, 2, 0): CFName("sea_ice_area_fraction", None, "1"),
    G2Param(2, 10, 3, 0): CFName("sea_surface_temperature", None, "K"),
}

CF_CONSTRAINED_TO_GRIB1_LOCAL = {
    (
        CFName("air_temperature", None, "K"),
        DimensionCoordinate("height", "m", (2,)),
    ): G1LocalParam(1, 128, 98, 167),
    (
        CFName("dew_point_temperature", None, "K"),
        DimensionCoordinate("height", "m", (2,)),
    ): G1LocalParam(1, 128, 98, 168),
    (
        CFName("x_wind", None, "m s-1"),
        DimensionCoordinate("height", "m", (10,)),
    ): G1LocalParam(1, 128, 98, 165),
    (
        CFName("y_wind", None, "m s-1"),
        DimensionCoordinate("height", "m", (10,)),
    ): G1LocalParam(1, 128, 98, 166),
}

CF_TO_GRIB1_LOCAL = {
    CFName(None, "grib_physical_atmosphere_albedo", "1"): G1LocalParam(1, 128, 98, 174),
    CFName(None, "grib_skin_temperature", "K"): G1LocalParam(1, 128, 98, 235),
    CFName("air_pressure_at_sea_level", None, "Pa"): G1LocalParam(1, 128, 98, 151),
    CFName("air_temperature", None, "K"): G1LocalParam(1, 128, 98, 130),
    CFName(
        "atmosphere_specific_convective_available_potential_energy", None, "J kg-1"
    ): G1LocalParam(1, 128, 98, 59),
    CFName("cloud_area_fraction", None, "1"): G1LocalParam(1, 128, 98, 164),
    CFName("geopotential", None, "m2 s-2"): G1LocalParam(1, 128, 98, 129),
    CFName("high_type_cloud_area_fraction", None, "1"): G1LocalParam(1, 128, 98, 188),
    CFName("lagrangian_tendency_of_air_pressure", None, "Pa s-1"): G1LocalParam(
        1, 128, 98, 135
    ),
    CFName("low_type_cloud_area_fraction", None, "1"): G1LocalParam(1, 128, 98, 186),
    CFName("medium_type_cloud_area_fraction", None, "1"): G1LocalParam(1, 128, 98, 187),
    CFName("relative_humidity", None, "%"): G1LocalParam(1, 128, 98, 157),
    CFName("sea_ice_area_fraction", None, "1"): G1LocalParam(1, 128, 98, 31),
    CFName("sea_surface_temperature", None, "K"): G1LocalParam(1, 128, 98, 34),
    CFName("surface_roughness_length", None, "m"): G1LocalParam(1, 128, 98, 173),
    CFName("thickness_of_snowfall_amount", None, "m"): G1LocalParam(1, 128, 98, 141),
    CFName("x_wind", None, "m s-1"): G1LocalParam(1, 128, 98, 131),
    CFName("y_wind", None, "m s-1"): G1LocalParam(1, 128, 98, 132),
}

CF_TO_GRIB2 = {
    CFName(None, "WAFC_CAT_potential", "1"): G2Param(2, 0, 19, 22),
    CFName(None, "WAFC_CB_horizontal_extent", "1"): G2Param(2, 0, 6, 25),
    CFName(None, "WAFC_ICAO_height_at_cloud_base", "m"): G2Param(2, 0, 6, 26),
    CFName(None, "WAFC_ICAO_height_at_cloud_top", "m"): G2Param(2, 0, 6, 27),
    CFName(None, "WAFC_icing_potential", "1"): G2Param(2, 0, 19, 20),
    CFName(None, "WAFC_in-cloud_turb_potential", "1"): G2Param(2, 0, 19, 21),
    CFName(None, "cloud_area_fraction_assuming_maximum_random_overlap", "1"): G2Param(
        2, 0, 6, 1
    ),
    CFName(None, "cloud_mixing_ratio", "kg kg-1"): G2Param(2, 0, 1, 22),
    CFName(None, "convective_inhibition", "J kg-1"): G2Param(2, 0, 7, 7),
    CFName(None, "ertel_potential_velocity", "K m2 kg-1 s-1"): G2Param(2, 0, 2, 14),
    CFName(None, "grib_physical_atmosphere_albedo", "%"): G2Param(2, 0, 19, 1),
    CFName(None, "icao_standard_atmosphere_reference_height", "m"): G2Param(2, 0, 3, 3),
    CFName(None, "precipitable_water", "kg m-2"): G2Param(2, 0, 1, 3),
    CFName(None, "storm_relative_helicity", "J kg-1"): G2Param(2, 0, 7, 8),
    CFName("air_potential_temperature", None, "K"): G2Param(2, 0, 0, 2),
    CFName("air_pressure", None, "Pa"): G2Param(2, 0, 3, 0),
    CFName("air_pressure_at_sea_level", None, "Pa"): G2Param(2, 0, 3, 0),
    CFName("air_pressure_at_sea_level", None, "Pa"): G2Param(2, 0, 3, 1),
    CFName("air_temperature", None, "K"): G2Param(2, 0, 0, 0),
    CFName("altitude", None, "m"): G2Param(2, 0, 3, 6),
    CFName("atmosphere_absolute_vorticity", None, "s-1"): G2Param(2, 0, 2, 10),
    CFName("atmosphere_mass_content_of_cloud_liquid_water", None, "kg m-2"): G2Param(
        2, 0, 6, 6
    ),
    CFName("atmosphere_mass_content_of_water", None, "kg m-2"): G2Param(2, 0, 1, 51),
    CFName("atmosphere_mass_content_of_water_vapor", None, "kg m-2"): G2Param(
        2, 0, 1, 64
    ),
    CFName("atmosphere_mole_content_of_ozone", None, "Dobson"): G2Param(2, 0, 14, 0),
    CFName("atmosphere_relative_vorticity", None, "s-1"): G2Param(2, 0, 2, 12),
    CFName(
        "atmosphere_specific_convective_available_potential_energy", None, "J kg-1"
    ): G2Param(2, 0, 7, 6),
    CFName("convective_rainfall_amount", None, "kg m-2"): G2Param(2, 0, 1, 10),
    CFName("convective_rainfall_flux", None, "kg m-2 s-1"): G2Param(2, 0, 1, 37),
    CFName("convective_snowfall_flux", None, "kg m-2 s-1"): G2Param(2, 0, 1, 58),
    CFName("cloud_area_fraction_in_atmosphere_layer", None, "%"): G2Param(2, 0, 6, 7),
    CFName("dew_point_temperature", None, "K"): G2Param(2, 0, 0, 6),
    CFName("geopotential", None, "m2 s-2"): G2Param(2, 0, 3, 4),
    CFName("geopotential_height", None, "m"): G2Param(2, 0, 3, 5),
    CFName("geopotential_height_anomaly", None, "m"): G2Param(2, 0, 3, 9),
    CFName("high_type_cloud_area_fraction", None, "%"): G2Param(2, 0, 6, 5),
    CFName("humidity_mixing_ratio", None, "kg kg-1"): G2Param(2, 0, 1, 2),
    CFName("lagrangian_tendency_of_air_pressure", None, "Pa s-1"): G2Param(2, 0, 2, 8),
    CFName("land_area_fraction", None, "1"): G2Param(2, 2, 0, 0),
    CFName("land_binary_mask", None, "1"): G2Param(2, 2, 0, 0),
    CFName("liquid_water_content_of_surface_snow", None, "kg m-2"): G2Param(
        2, 0, 1, 13
    ),
    CFName("low_type_cloud_area_fraction", None, "%"): G2Param(2, 0, 6, 3),
    CFName("mass_fraction_of_cloud_ice_in_air", None, "kg kg-1"): G2Param(2, 0, 1, 84),
    CFName("mass_fraction_of_cloud_liquid_water_in_air", None, "kg kg-1"): G2Param(
        2, 0, 1, 83
    ),
    CFName("medium_type_cloud_area_fraction", None, "%"): G2Param(2, 0, 6, 4),
    CFName("moisture_content_of_soil_layer", None, "kg m-2"): G2Param(2, 2, 0, 22),
    CFName("precipitation_amount", None, "kg m-2"): G2Param(2, 0, 1, 49),
    CFName("precipitation_flux", None, "kg m-2 s-1"): G2Param(2, 0, 1, 7),
    CFName("relative_humidity", None, "%"): G2Param(2, 0, 1, 1),
    CFName("sea_ice_area_fraction", None, "1"): G2Param(2, 10, 2, 0),
    CFName("sea_surface_temperature", None, "K"): G2Param(2, 10, 3, 0),
    CFName("sea_water_x_velocity", None, "m s-1"): G2Param(2, 10, 1, 2),
    CFName("sea_water_y_velocity", None, "m s-1"): G2Param(2, 10, 1, 3),
    CFName("snowfall_amount", None, "kg m-2"): G2Param(2, 0, 1, 60),
    CFName("snowfall_flux", None, "kg m-2 s-1"): G2Param(2, 0, 1, 53),
    CFName("soil_moisture_content", None, "kg m-2"): G2Param(2, 2, 0, 3),
    CFName("soil_temperature", None, "K"): G2Param(2, 2, 0, 2),
    CFName("specific_humidity", None, "kg kg-1"): G2Param(2, 0, 1, 0),
    CFName("stratiform_rainfall_amount", None, "kg m-2"): G2Param(2, 0, 1, 9),
    CFName("stratiform_rainfall_flux", None, "kg m-2 s-1"): G2Param(2, 0, 1, 77),
    CFName("stratiform_snowfall_amount", None, "kg m-2"): G2Param(2, 0, 1, 15),
    CFName("stratiform_snowfall_flux", None, "kg m-2 s-1"): G2Param(2, 0, 1, 59),
    CFName("surface_air_pressure", None, "Pa"): G2Param(2, 0, 3, 0),
    CFName("surface_altitude", None, "m"): G2Param(2, 2, 0, 7),
    CFName("surface_downwelling_longwave_flux_in_air", None, "W m-2"): G2Param(
        2, 0, 5, 3
    ),
    CFName("surface_downwelling_shortwave_flux_in_air", None, "W m-2"): G2Param(
        2, 0, 4, 7
    ),
    CFName("surface_net_downward_longwave_flux", None, "W m-2"): G2Param(2, 0, 5, 5),
    CFName("surface_net_downward_shortwave_flux", None, "W m-2"): G2Param(2, 0, 4, 9),
    CFName("surface_roughness_length", None, "m"): G2Param(2, 2, 0, 1),
    CFName("surface_runoff_flux", None, "kg m-2 s-1"): G2Param(2, 2, 0, 34),
    CFName("surface_temperature", None, "K"): G2Param(2, 0, 0, 17),
    CFName("surface_upward_latent_heat_flux", None, "W m-2"): G2Param(2, 0, 0, 10),
    CFName("surface_upward_sensible_heat_flux", None, "W m-2"): G2Param(2, 0, 0, 11),
    CFName("toa_outgoing_longwave_flux", None, "W m-2"): G2Param(2, 0, 5, 4),
    CFName("thickness_of_snowfall_amount", None, "m"): G2Param(2, 0, 1, 11),
    CFName("upward_air_velocity", None, "m s-1"): G2Param(2, 0, 2, 9),
    CFName("wet_bulb_potential_temperature", None, "K"): G2Param(2, 0, 0, 32),
    CFName("wind_from_direction", None, "degrees"): G2Param(2, 0, 2, 0),
    CFName("wind_speed", None, "m s-1"): G2Param(2, 0, 2, 1),
    CFName("wind_speed_of_gust", None, "m s-1"): G2Param(2, 0, 2, 22),
    CFName("x_wind", None, "m s-1"): G2Param(2, 0, 2, 2),
    CFName("y_wind", None, "m s-1"): G2Param(2, 0, 2, 3),
}



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.product_definition_template_40`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import iris.coords

from iris_grib._load_convert import product_definition_template_40, _MDI
from iris_grib.tests.unit.load_convert import empty_metadata


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section_4 = {
            "hoursAfterDataCutoff": _MDI,
            "minutesAfterDataCutoff": _MDI,
            "constituentType": 1,
            "indicatorOfUnitForForecastTime": 0,  # minutes
            "startStep": 360,
            "NV": 0,
            "typeOfFirstFixedSurface": 103,
            "scaleFactorOfFirstFixedSurface": 0,
            "scaledValueOfFirstFixedSurface": 9999,
            "typeOfSecondFixedSurface": 255,
        }

    def test_constituent_type(self):
        metadata = empty_metadata()
        rt_coord = iris.coords.DimCoord(
            24, "forecast_reference_time", units="hours since epoch"
        )
        product_definition_template_40(self.section_4, metadata, rt_coord)
        expected = empty_metadata()
        expected["attributes"]["WMO_constituent_type"] = 1
        self.assertEqual(metadata["attributes"], expected["attributes"])


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""Round-trip integration tests for the :mod:`iris-grib` package."""



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function
:func:`iris_grib._load_convert.product_definition_template_1`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import DimCoord

from iris_grib._load_convert import product_definition_template_1


class Test(tests.IrisGribTest):
    def setUp(self):
        def func(s, m, f):
            return m["cell_methods"].append(self.cell_method)

        module = "iris_grib._load_convert"
        self.patch("warnings.warn")
        this = "{}.product_definition_template_0".format(module)
        self.cell_method = mock.sentinel.cell_method
        self.patch(this, side_effect=func)
        self.metadata = {
            "factories": [],
            "references": [],
            "standard_name": None,
            "long_name": None,
            "units": None,
            "attributes": {},
            "cell_methods": [],
            "dim_coords_and_dims": [],
            "aux_coords_and_dims": [],
        }

    def _check(self, request_warning):
        this = "iris_grib._load_convert.options"
        with mock.patch(this, warn_on_unsupported=request_warning):
            metadata = deepcopy(self.metadata)
            perturbationNumber = 666
            section = {"perturbationNumber": perturbationNumber}
            forecast_reference_time = mock.sentinel.forecast_reference_time
            # The called being tested.
            product_definition_template_1(section, metadata, forecast_reference_time)
            expected = deepcopy(self.metadata)
            expected["cell_methods"].append(self.cell_method)
            realization = DimCoord(
                perturbationNumber, standard_name="realization", units="no_unit"
            )
            expected["aux_coords_and_dims"].append((realization, None))
            self.assertEqual(metadata, expected)
            if request_warning:
                warn_msgs = [mcall[1][0] for mcall in warnings.warn.mock_calls]
                expected_msgs = ["type of ensemble", "number of forecasts"]
                for emsg in expected_msgs:
                    matches = [wmsg for wmsg in warn_msgs if emsg in wmsg]
                    self.assertEqual(len(matches), 1)
                    warn_msgs.remove(matches[0])
            else:
                self.assertEqual(len(warnings.warn.mock_calls), 0)

    def test_pdt_no_warn(self):
        self._check(False)

    def test_pdt_warn(self):
        self._check(True)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function
:func:`iris_grib._load_convert.grid_definition_template_5`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock

from iris_grib._load_convert import grid_definition_template_5


class Test(tests.IrisGribTest):
    def setUp(self):
        def func(s, m, y, x, c):
            return m["dim_coords_and_dims"].append(item)

        module = "iris_grib._load_convert"

        self.major = mock.sentinel.major
        self.minor = mock.sentinel.minor
        self.radius = mock.sentinel.radius

        mfunc = "{}.ellipsoid_geometry".format(module)
        return_value = (self.major, self.minor, self.radius)
        self.patch(mfunc, return_value=return_value)

        mfunc = "{}.ellipsoid".format(module)
        self.ellipsoid = mock.sentinel.ellipsoid
        self.patch(mfunc, return_value=self.ellipsoid)

        mfunc = "{}.grid_definition_template_4_and_5".format(module)
        self.coord = mock.sentinel.coord
        self.dim = mock.sentinel.dim
        item = (self.coord, self.dim)
        self.patch(mfunc, side_effect=func)

        mclass = "iris.coord_systems.RotatedGeogCS"
        self.cs = mock.sentinel.cs
        self.patch(mclass, return_value=self.cs)

        self.metadata = {
            "factories": [],
            "references": [],
            "standard_name": None,
            "long_name": None,
            "units": None,
            "attributes": {},
            "cell_methods": [],
            "dim_coords_and_dims": [],
            "aux_coords_and_dims": [],
        }

    def test(self):
        metadata = deepcopy(self.metadata)
        angleOfRotation = mock.sentinel.angleOfRotation
        shapeOfTheEarth = mock.sentinel.shapeOfTheEarth
        section = {
            "latitudeOfSouthernPole": 45000000,
            "longitudeOfSouthernPole": 90000000,
            "angleOfRotation": angleOfRotation,
            "shapeOfTheEarth": shapeOfTheEarth,
        }
        # The called being tested.
        grid_definition_template_5(section, metadata)
        from iris_grib._load_convert import (
            ellipsoid_geometry,
            ellipsoid,
            grid_definition_template_4_and_5 as gdt_4_5,
        )

        self.assertEqual(ellipsoid_geometry.call_count, 1)
        ellipsoid.assert_called_once_with(
            shapeOfTheEarth, self.major, self.minor, self.radius
        )
        from iris.coord_systems import RotatedGeogCS

        RotatedGeogCS.assert_called_once_with(
            -45.0, 270.0, angleOfRotation, self.ellipsoid
        )
        gdt_4_5.assert_called_once_with(
            section, metadata, "grid_latitude", "grid_longitude", self.cs
        )
        expected = deepcopy(self.metadata)
        expected["dim_coords_and_dims"].append((self.coord, self.dim))
        self.assertEqual(metadata, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.product_definition_template_10`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris.coords import DimCoord
from iris_grib._load_convert import product_definition_template_10
from iris_grib.tests.unit.load_convert import empty_metadata


class Test(tests.IrisGribTest):
    def setUp(self):
        module = "iris_grib._load_convert"
        self.patch_statistical_fp_coord = self.patch(
            module + ".statistical_forecast_period_coord",
            return_value=mock.sentinel.dummy_fp_coord,
        )
        self.patch_time_coord = self.patch(
            module + ".validity_time_coord", return_value=mock.sentinel.dummy_time_coord
        )
        self.patch_vertical_coords = self.patch(module + ".vertical_coords")

    def test_percentile_coord(self):
        metadata = empty_metadata()
        percentileValue = 75
        section = {
            "productDefinitionTemplateNumber": 10,
            "percentileValue": percentileValue,
            "hoursAfterDataCutoff": 1,
            "minutesAfterDataCutoff": 1,
            "numberOfTimeRange": 1,
            "typeOfStatisticalProcessing": 1,
            "typeOfTimeIncrement": 2,
            "timeIncrement": 0,
            "yearOfEndOfOverallTimeInterval": 2000,
            "monthOfEndOfOverallTimeInterval": 1,
            "dayOfEndOfOverallTimeInterval": 1,
            "hourOfEndOfOverallTimeInterval": 1,
            "minuteOfEndOfOverallTimeInterval": 0,
            "secondOfEndOfOverallTimeInterval": 1,
        }
        forecast_reference_time = mock.Mock()
        # The called being tested.
        product_definition_template_10(section, metadata, forecast_reference_time)

        expected = {"aux_coords_and_dims": []}
        percentile = DimCoord(
            percentileValue, long_name="percentile_over_time", units="no_unit"
        )
        expected["aux_coords_and_dims"].append((percentile, None))

        self.assertEqual(
            metadata["aux_coords_and_dims"][-1], expected["aux_coords_and_dims"][0]
        )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.other_time_coord.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import iris.coords

from iris_grib._load_convert import other_time_coord


class TestValid(tests.IrisGribTest):
    def test_t(self):
        rt = iris.coords.DimCoord(48, "time", units="hours since epoch")
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        result = other_time_coord(rt, fp)
        expected = iris.coords.DimCoord(
            42, "forecast_reference_time", units="hours since epoch"
        )
        self.assertEqual(result, expected)

    def test_frt(self):
        rt = iris.coords.DimCoord(
            48, "forecast_reference_time", units="hours since epoch"
        )
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        result = other_time_coord(rt, fp)
        expected = iris.coords.DimCoord(54, "time", units="hours since epoch")
        self.assertEqual(result, expected)


class TestInvalid(tests.IrisGribTest):
    def test_t_with_bounds(self):
        rt = iris.coords.DimCoord(
            48, "time", units="hours since epoch", bounds=[36, 60]
        )
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "bounds"):
            other_time_coord(rt, fp)

    def test_frt_with_bounds(self):
        rt = iris.coords.DimCoord(
            48, "forecast_reference_time", units="hours since epoch", bounds=[42, 54]
        )
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "bounds"):
            other_time_coord(rt, fp)

    def test_fp_with_bounds(self):
        rt = iris.coords.DimCoord(48, "time", units="hours since epoch")
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours", bounds=[3, 9])
        with self.assertRaisesRegex(ValueError, "bounds"):
            other_time_coord(rt, fp)

    def test_vector_t(self):
        rt = iris.coords.DimCoord([0, 3], "time", units="hours since epoch")
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "Vector"):
            other_time_coord(rt, fp)

    def test_vector_frt(self):
        rt = iris.coords.DimCoord(
            [0, 3], "forecast_reference_time", units="hours since epoch"
        )
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "Vector"):
            other_time_coord(rt, fp)

    def test_vector_fp(self):
        rt = iris.coords.DimCoord(48, "time", units="hours since epoch")
        fp = iris.coords.DimCoord([6, 12], "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "Vector"):
            other_time_coord(rt, fp)

    def test_invalid_rt_name(self):
        rt = iris.coords.DimCoord(1, "height")
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "reference time"):
            other_time_coord(rt, fp)

    def test_invalid_t_unit(self):
        rt = iris.coords.DimCoord(1, "time", units="Pa")
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "unit.*Pa"):
            other_time_coord(rt, fp)

    def test_invalid_frt_unit(self):
        rt = iris.coords.DimCoord(1, "forecast_reference_time", units="km")
        fp = iris.coords.DimCoord(6, "forecast_period", units="hours")
        with self.assertRaisesRegex(ValueError, "unit.*km"):
            other_time_coord(rt, fp)

    def test_invalid_fp_unit(self):
        rt = iris.coords.DimCoord(
            48, "forecast_reference_time", units="hours since epoch"
        )
        fp = iris.coords.DimCoord(6, "forecast_period", units="kg")
        with self.assertRaisesRegex(ValueError, "unit.*kg"):
            other_time_coord(rt, fp)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for function
:func:`iris_grib._load_convert.product_definition_template_8`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris_grib._load_convert import product_definition_template_8


class Test(tests.IrisGribTest):
    def setUp(self):
        module = "iris_grib._load_convert"
        self.module = module
        # Create patches for called routines
        self.patch_generating_process = self.patch(module + ".generating_process")
        self.patch_data_cutoff = self.patch(module + ".data_cutoff")
        self.patch_statistical_cell_method = self.patch(
            module + ".statistical_cell_method",
            return_value=mock.sentinel.dummy_cell_method,
        )
        self.patch_statistical_fp_coord = self.patch(
            module + ".statistical_forecast_period_coord",
            return_value=mock.sentinel.dummy_fp_coord,
        )
        self.patch_time_coord = self.patch(
            module + ".validity_time_coord", return_value=mock.sentinel.dummy_time_coord
        )
        self.patch_vertical_coords = self.patch(module + ".vertical_coords")
        # Construct dummy call arguments
        self.section = {}
        self.section["hoursAfterDataCutoff"] = mock.sentinel.cutoff_hours
        self.section["minutesAfterDataCutoff"] = mock.sentinel.cutoff_mins
        self.frt_coord = mock.Mock()
        self.metadata = {"cell_methods": [], "aux_coords_and_dims": []}

    def test_basic(self):
        product_definition_template_8(self.section, self.metadata, self.frt_coord)
        # Check all expected functions were called just once.
        self.assertEqual(self.patch_generating_process.call_count, 1)
        self.assertEqual(self.patch_data_cutoff.call_count, 1)
        self.assertEqual(self.patch_statistical_cell_method.call_count, 1)
        self.assertEqual(self.patch_statistical_fp_coord.call_count, 1)
        self.assertEqual(self.patch_time_coord.call_count, 1)
        self.assertEqual(self.patch_vertical_coords.call_count, 1)
        # Check metadata content.
        self.assertEqual(
            sorted(self.metadata.keys()), ["aux_coords_and_dims", "cell_methods"]
        )
        self.assertEqual(
            self.metadata["cell_methods"], [mock.sentinel.dummy_cell_method]
        )
        self.assertCountEqual(
            self.metadata["aux_coords_and_dims"],
            [
                (self.frt_coord, None),
                (mock.sentinel.dummy_fp_coord, None),
                (mock.sentinel.dummy_time_coord, None),
            ],
        )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for `iris_grib._load_convert.satellite_common`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris.coords import AuxCoord
import numpy as np

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import satellite_common


class Test(tests.IrisGribTest):
    def _check(self, factors=1, values=111):
        # Prepare the arguments.
        series = mock.sentinel.satelliteSeries
        number = mock.sentinel.satelliteNumber
        instrument = mock.sentinel.instrumentType
        section = {
            "NB": 1,
            "satelliteSeries": series,
            "satelliteNumber": number,
            "instrumentType": instrument,
            "scaleFactorOfCentralWaveNumber": factors,
            "scaledValueOfCentralWaveNumber": values,
        }

        # Call the function.
        metadata = empty_metadata()
        satellite_common(section, metadata)

        # Check the result.
        expected = empty_metadata()
        coord = AuxCoord(series, long_name="satellite_series", units=1)
        expected["aux_coords_and_dims"].append((coord, None))
        coord = AuxCoord(number, long_name="satellite_number", units=1)
        expected["aux_coords_and_dims"].append((coord, None))
        coord = AuxCoord(instrument, long_name="instrument_type", units=1)
        expected["aux_coords_and_dims"].append((coord, None))
        standard_name = "sensor_band_central_radiation_wavenumber"
        coord = AuxCoord(
            values / (10.0**factors), standard_name=standard_name, units="m-1"
        )
        expected["aux_coords_and_dims"].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_basic(self):
        self._check()

    def test_multiple_wavelengths(self):
        # Check with multiple values, and several different scaling factors.
        values = np.array([1, 11, 123, 1975])
        for i_factor in (-3, -1, 0, 1, 3):
            factors = np.ones(values.shape) * i_factor
            self._check(values=values, factors=factors)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function
:func:`iris_grib._load_convert.grid_definition_template_4_and_5`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock
import warnings

from iris.coords import DimCoord
import numpy as np

from iris_grib._load_convert import grid_definition_template_4_and_5, _MDI as MDI


RESOLUTION = 1e6


class Test(tests.IrisGribTest):
    def setUp(self):
        self.patch("warnings.warn")
        self.patch("iris_grib._load_convert._is_circular", return_value=False)
        self.metadata = {
            "factories": [],
            "references": [],
            "standard_name": None,
            "long_name": None,
            "units": None,
            "attributes": {},
            "cell_methods": [],
            "dim_coords_and_dims": [],
            "aux_coords_and_dims": [],
        }
        self.cs = mock.sentinel.coord_system
        self.data = np.arange(10, dtype=np.float64)

    def _check(self, section, request_warning, expect_warning=False, y_dim=0, x_dim=1):
        this = "iris_grib._load_convert.options"
        with mock.patch(this, warn_on_unsupported=request_warning):
            metadata = deepcopy(self.metadata)
            # The called being tested.
            grid_definition_template_4_and_5(
                section, metadata, "latitude", "longitude", self.cs
            )
            expected = deepcopy(self.metadata)
            coord = DimCoord(
                self.data,
                standard_name="latitude",
                units="degrees",
                coord_system=self.cs,
            )
            expected["dim_coords_and_dims"].append((coord, y_dim))
            coord = DimCoord(
                self.data,
                standard_name="longitude",
                units="degrees",
                coord_system=self.cs,
            )
            expected["dim_coords_and_dims"].append((coord, x_dim))
            self.assertEqual(metadata, expected)
            if expect_warning:
                self.assertEqual(len(warnings.warn.mock_calls), 1)
                args, kwargs = warnings.warn.call_args
                self.assertIn("resolution and component flags", args[0])
            else:
                self.assertEqual(len(warnings.warn.mock_calls), 0)

    def test_resolution_default_0(self):
        for request_warn in [False, True]:
            section = {
                "basicAngleOfTheInitialProductionDomain": 0,
                "subdivisionsOfBasicAngle": 0,
                "resolutionAndComponentFlags": 0,
                "longitudes": self.data * RESOLUTION,
                "latitudes": self.data * RESOLUTION,
                "scanningMode": 0,
            }
            self._check(section, request_warn)

    def test_resolution_default_mdi(self):
        for request_warn in [False, True]:
            section = {
                "basicAngleOfTheInitialProductionDomain": MDI,
                "subdivisionsOfBasicAngle": MDI,
                "resolutionAndComponentFlags": 0,
                "longitudes": self.data * RESOLUTION,
                "latitudes": self.data * RESOLUTION,
                "scanningMode": 0,
            }
            self._check(section, request_warn)

    def test_resolution(self):
        angle = 10
        for request_warn in [False, True]:
            section = {
                "basicAngleOfTheInitialProductionDomain": 1,
                "subdivisionsOfBasicAngle": angle,
                "resolutionAndComponentFlags": 0,
                "longitudes": self.data * angle,
                "latitudes": self.data * angle,
                "scanningMode": 0,
            }
            self._check(section, request_warn)

    def test_uv_resolved_warn(self):
        angle = 100
        for warn in [False, True]:
            section = {
                "basicAngleOfTheInitialProductionDomain": 1,
                "subdivisionsOfBasicAngle": angle,
                "resolutionAndComponentFlags": 0x08,
                "longitudes": self.data * angle,
                "latitudes": self.data * angle,
                "scanningMode": 0,
            }
            self._check(section, warn, expect_warning=warn)

    def test_j_consecutive(self):
        angle = 1000
        for request_warn in [False, True]:
            section = {
                "basicAngleOfTheInitialProductionDomain": 1,
                "subdivisionsOfBasicAngle": angle,
                "resolutionAndComponentFlags": 0,
                "longitudes": self.data * angle,
                "latitudes": self.data * angle,
                "scanningMode": 0x20,
            }
            self._check(section, request_warn, y_dim=1, x_dim=0)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for function
:func:`iris_grib._load_convert.product_definition_template_9`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris.exceptions import TranslationError

from iris_grib._load_convert import product_definition_template_9
from iris_grib._load_convert import Probability, _MDI


class Test(tests.IrisGribTest):
    def setUp(self):
        # Create patches for called routines
        module = "iris_grib._load_convert"
        self.patch_pdt8_call = self.patch(module + ".product_definition_template_8")
        # Construct dummy call arguments
        self.section = {}
        self.section["probabilityType"] = 1
        self.section["scaledValueOfUpperLimit"] = 53
        self.section["scaleFactorOfUpperLimit"] = 1
        self.frt_coord = mock.sentinel.frt_coord
        self.metadata = {
            "cell_methods": [mock.sentinel.cell_method],
            "aux_coords_and_dims": [],
        }

    def test_basic(self):
        result = product_definition_template_9(
            self.section, self.metadata, self.frt_coord
        )
        # Check expected function was called.
        self.assertEqual(
            self.patch_pdt8_call.call_args_list,
            [mock.call(self.section, self.metadata, self.frt_coord)],
        )
        # Check metadata content (N.B. cell_method has been removed!).
        self.assertEqual(self.metadata, {"cell_methods": [], "aux_coords_and_dims": []})
        # Check result.
        self.assertEqual(result, Probability("above_threshold", 5.3))

    def test_fail_bad_probability_type(self):
        self.section["probabilityType"] = 17
        with self.assertRaisesRegex(TranslationError, "unsupported probability type"):
            product_definition_template_9(self.section, self.metadata, self.frt_coord)

    def test_fail_bad_threshold_value(self):
        self.section["scaledValueOfUpperLimit"] = _MDI
        with self.assertRaisesRegex(TranslationError, "missing scaled value"):
            product_definition_template_9(self.section, self.metadata, self.frt_coord)

    def test_fail_bad_threshold_scalefactor(self):
        self.section["scaleFactorOfUpperLimit"] = _MDI
        with self.assertRaisesRegex(TranslationError, "missing scale factor"):
            product_definition_template_9(self.section, self.metadata, self.frt_coord)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Integration test for loading hybrid height data.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests


from iris import load_cube
from iris.aux_factory import HybridHeightFactory, HybridPressureFactory


@tests.skip_grib_data
class TestHybridHeight(tests.IrisGribTest):
    def setUp(self):
        filepath = self.get_testdata_path("faked_sample_hh_grib_data.grib2")
        self.testdata_cube = load_cube(filepath, "air_temperature")

    def test_load_hybrid_height(self):
        # Check that it loads right, and creates a factory.
        self.assertIsInstance(self.testdata_cube.aux_factories[0], HybridHeightFactory)

    def test_hybrid_height_coords_values(self):
        cube = self.testdata_cube

        # check actual model level values.
        self.assertArrayEqual(cube.coord("model_level_number").points, [1, 11, 21])

        # check sigma values correctly loaded from indices 1, 11, 21.
        self.assertArrayAllClose(
            cube.coord("sigma").points, [0.998, 0.894, 0.667], atol=0.001
        )

        # check height values too.
        self.assertArrayAllClose(
            cube.coord("level_height").points, [20.0, 953.3, 3220.0], atol=0.5
        )


@tests.skip_grib_data
class TestHybridPressure(tests.IrisGribTest):
    def setUp(self):
        filepath = self.get_testdata_path("faked_sample_hp_grib_data.grib2")
        self.testdata_cube = load_cube(filepath, "air_temperature")

    def test_load_hybrid_pressure(self):
        # Check that it loads right, and creates a factory.
        self.assertIsInstance(
            self.testdata_cube.aux_factories[0], HybridPressureFactory
        )

    def test_hybrid_pressure_coords_values(self):
        cube = self.testdata_cube

        # Check existence, and some values, of the loaded coefficients.
        self.assertArrayEqual(cube.coord("model_level_number").points, [1, 51, 91])
        self.assertArrayAllClose(
            cube.coord("sigma").points, [0.0, 0.045, 1.0], atol=0.001
        )
        self.assertArrayAllClose(
            cube.coord("level_pressure").points, [2.00004, 18716.9688, 0.0], rtol=0.0001
        )
        self.assertArrayAllClose(
            cube.coord("surface_air_pressure")[:2, :3].points,
            [[103493.8, 103493.8, 103493.8], [103401.0, 103407.4, 103412.2]],
            atol=0.1,
        )

        # Also check a few values from the derived coord.
        self.assertArrayAllClose(
            cube.coord("air_pressure")[:, :3, 0].points,
            [
                [2.0, 2.0, 2.0],
                [23389.3, 23385.1, 23379.9],
                [103493.8, 103401.0, 103285.8],
            ],
            atol=0.1,
        )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.

# Historically this was auto-generated from
# SciTools/iris-code-generators:tools/gen_rules.py


from cf_units import CALENDAR_GREGORIAN, Unit

from iris.aux_factory import HybridPressureFactory
from iris.coords import AuxCoord, CellMethod, DimCoord
from iris.exceptions import TranslationError
from iris.fileformats.rules import ConversionMetadata, Factory, Reference


def grib1_convert(grib):
    """
    Converts a GRIB1 message into the corresponding items of Cube metadata.

    Args:

    * grib:
        A :class:`~iris_grib.GribWrapper` object.

    Returns:
        A :class:`iris.fileformats.rules.ConversionMetadata` object.

    """
    if grib.edition != 1:
        emsg = "GRIB edition {} is not supported by {!r}."
        raise TranslationError(emsg.format(grib.edition, type(grib).__name__))

    factories = []
    references = []
    standard_name = None
    long_name = None
    units = None
    attributes = {}
    cell_methods = []
    dim_coords_and_dims = []
    aux_coords_and_dims = []

    if grib.gridType == "reduced_gg":
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    grib._y_points,
                    grib._y_coord_name,
                    units="degrees",
                    coord_system=grib._coord_system,
                ),
                0,
            )
        )
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    grib._x_points,
                    grib._x_coord_name,
                    units="degrees",
                    coord_system=grib._coord_system,
                ),
                0,
            )
        )
    elif grib.gridType in ("regular_ll", "rotated_ll", "regular_gg"):
        j_points_are_consecutive = grib.jPointsAreConsecutive

        if j_points_are_consecutive not in (0, 1):
            raise ValueError

        dim_coords_and_dims.append(
            (
                DimCoord(
                    grib._y_points,
                    grib._y_coord_name,
                    units="degrees",
                    coord_system=grib._coord_system,
                ),
                j_points_are_consecutive,
            )
        )
        dim_coords_and_dims.append(
            (
                DimCoord(
                    grib._x_points,
                    grib._x_coord_name,
                    units="degrees",
                    coord_system=grib._coord_system,
                    circular=grib._x_circular,
                ),
                int(not j_points_are_consecutive),
            )
        )

    elif grib.gridType in ["polar_stereographic", "lambert"]:
        dim_coords_and_dims.append(
            (
                DimCoord(
                    grib._y_points,
                    grib._y_coord_name,
                    units="m",
                    coord_system=grib._coord_system,
                ),
                0,
            )
        )
        dim_coords_and_dims.append(
            (
                DimCoord(
                    grib._x_points,
                    grib._x_coord_name,
                    units="m",
                    coord_system=grib._coord_system,
                ),
                1,
            )
        )

    if (
        (grib.table2Version < 128)
        and (grib.indicatorOfParameter == 11)
        and (grib._cf_data is None)
    ):
        standard_name = "air_temperature"
        units = "kelvin"

    if (
        (grib.table2Version < 128)
        and (grib.indicatorOfParameter == 33)
        and (grib._cf_data is None)
    ):
        standard_name = "x_wind"
        units = "m s-1"

    if (
        (grib.table2Version < 128)
        and (grib.indicatorOfParameter == 34)
        and (grib._cf_data is None)
    ):
        standard_name = "y_wind"
        units = "m s-1"

    if grib._cf_data is not None:
        standard_name = grib._cf_data.standard_name
        long_name = grib._cf_data.standard_name or grib._cf_data.long_name
        units = grib._cf_data.units

    # N.B. in addition to the previous cf translated phenomenon info,
    # **always** add a GRIB_PARAM attribute to identify the input phenomenon
    # identity.
    attributes["GRIB_PARAM"] = grib._grib_code

    if (
        (grib.table2Version >= 128)
        and (grib._cf_data is None)
        or (grib.table2Version == 1)
        and (grib.indicatorOfParameter >= 128)
    ):
        long_name = (
            f"UNKNOWN LOCAL PARAM {grib.indicatorOfParameter}" f".{grib.table2Version}"
        )
        units = "???"

    if grib._phenomenonDateTime != -1.0:
        aux_coords_and_dims.append(
            (
                DimCoord(
                    points=grib.startStep,
                    standard_name="forecast_period",
                    units=grib._forecastTimeUnit,
                ),
                None,
            )
        )
        aux_coords_and_dims.append(
            (
                DimCoord(
                    points=grib.phenomenon_points("hours"),
                    standard_name="time",
                    units=Unit("hours since epoch", CALENDAR_GREGORIAN),
                ),
                None,
            )
        )

    def add_bounded_time_coords(aux_coords_and_dims, grib):
        t_bounds = grib.phenomenon_bounds("hours")
        period = t_bounds[1] - t_bounds[0]
        aux_coords_and_dims.append(
            (
                DimCoord(
                    standard_name="forecast_period",
                    units="hours",
                    points=grib._forecastTime + 0.5 * period,
                    bounds=[grib._forecastTime, grib._forecastTime + period],
                ),
                None,
            )
        )
        aux_coords_and_dims.append(
            (
                DimCoord(
                    standard_name="time",
                    units=Unit("hours since epoch", CALENDAR_GREGORIAN),
                    points=0.5 * (t_bounds[0] + t_bounds[1]),
                    bounds=t_bounds,
                ),
                None,
            )
        )

    if grib.timeRangeIndicator == 2:
        add_bounded_time_coords(aux_coords_and_dims, grib)

    if grib.timeRangeIndicator == 3:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if grib.timeRangeIndicator == 4:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if grib.timeRangeIndicator == 5:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("_difference", coords="time"))

    if grib.timeRangeIndicator == 51:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if grib.timeRangeIndicator == 113:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if grib.timeRangeIndicator == 114:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if grib.timeRangeIndicator == 115:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if grib.timeRangeIndicator == 116:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if grib.timeRangeIndicator == 117:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if grib.timeRangeIndicator == 118:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("_covariance", coords="time"))

    if grib.timeRangeIndicator == 123:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("mean", coords="time"))

    if grib.timeRangeIndicator == 124:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("sum", coords="time"))

    if grib.timeRangeIndicator == 125:
        add_bounded_time_coords(aux_coords_and_dims, grib)
        cell_methods.append(CellMethod("standard_deviation", coords="time"))

    if grib.levelType == "pl":
        aux_coords_and_dims.append(
            (DimCoord(points=grib.level, long_name="pressure", units="hPa"), None)
        )

    if grib.levelType == "sfc":
        if grib._cf_data is not None and grib._cf_data.set_height is not None:
            aux_coords_and_dims.append(
                (
                    DimCoord(
                        points=grib._cf_data.set_height,
                        long_name="height",
                        units="m",
                        attributes={"positive": "up"},
                    ),
                    None,
                )
            )
        elif grib.typeOfLevel == "heightAboveGround":  # required for NCAR
            aux_coords_and_dims.append(
                (
                    DimCoord(
                        points=grib.level,
                        long_name="height",
                        units="m",
                        attributes={"positive": "up"},
                    ),
                    None,
                )
            )

    if grib.levelType == "ml" and hasattr(grib, "pv"):
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    grib.level,
                    standard_name="model_level_number",
                    units=1,
                    attributes={"positive": "up"},
                ),
                None,
            )
        )
        aux_coords_and_dims.append(
            (
                DimCoord(grib.pv[grib.level], long_name="level_pressure", units="Pa"),
                None,
            )
        )
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    grib.pv[grib.numberOfCoordinatesValues // 2 + grib.level],
                    long_name="sigma",
                    units=1,
                ),
                None,
            )
        )
        factories.append(
            Factory(
                HybridPressureFactory,
                [
                    {"long_name": "level_pressure"},
                    {"long_name": "sigma"},
                    Reference("surface_pressure"),
                ],
            )
        )

    if grib._originatingCentre != "unknown":
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    points=grib._originatingCentre,
                    long_name="originating_centre",
                    units="no_unit",
                ),
                None,
            )
        )

    return ConversionMetadata(
        factories,
        references,
        standard_name,
        long_name,
        units,
        attributes,
        cell_methods,
        dim_coords_and_dims,
        aux_coords_and_dims,
    )



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Integration tests for format interoperability.

Checks that specific features of cubes loaded from other formats are correctly
represented by grib saving.

"""



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for `iris_grib._load_convert.product_definition_template_32`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import product_definition_template_32


class Test(tests.IrisGribTest):
    def setUp(self):
        self.patch("warnings.warn")
        self.generating_process_patch = self.patch(
            "iris_grib._load_convert.generating_process"
        )
        self.satellite_common_patch = self.patch(
            "iris_grib._load_convert.satellite_common"
        )
        self.time_coords_patch = self.patch("iris_grib._load_convert.time_coords")
        self.data_cutoff_patch = self.patch("iris_grib._load_convert.data_cutoff")

    def test(self, value=10, factor=1):
        # Prepare the arguments.
        series = mock.sentinel.satelliteSeries
        number = mock.sentinel.satelliteNumber
        instrument = mock.sentinel.instrumentType
        rt_coord = mock.sentinel.observation_time
        section = {
            "NB": 1,
            "hoursAfterDataCutoff": None,
            "minutesAfterDataCutoff": None,
            "satelliteSeries": series,
            "satelliteNumber": number,
            "instrumentType": instrument,
            "scaleFactorOfCentralWaveNumber": 1,
            "scaledValueOfCentralWaveNumber": 12,
        }

        # Call the function.
        metadata = empty_metadata()
        product_definition_template_32(section, metadata, rt_coord)

        # Check that 'satellite_common' was called.
        self.assertEqual(self.satellite_common_patch.call_count, 1)
        # Check that 'generating_process' was called.
        self.assertEqual(self.generating_process_patch.call_count, 1)
        # Check that 'data_cutoff' was called.
        self.assertEqual(self.data_cutoff_patch.call_count, 1)
        # Check that 'time_coords' was called.
        self.assertEqual(self.time_coords_patch.call_count, 1)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for
:func:`iris_grib._load_convert.grid_definition_template_12`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import numpy as np
import warnings

import iris.coord_systems
import iris.coords
import iris.exceptions

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import grid_definition_template_12


class Test(tests.IrisGribTest):
    def section_3(self):
        section = {
            "shapeOfTheEarth": 7,
            "scaleFactorOfRadiusOfSphericalEarth": MDI,
            "scaledValueOfRadiusOfSphericalEarth": MDI,
            "scaleFactorOfEarthMajorAxis": 3,
            "scaledValueOfEarthMajorAxis": 6377563396,
            "scaleFactorOfEarthMinorAxis": 3,
            "scaledValueOfEarthMinorAxis": 6356256909,
            "Ni": 4,
            "Nj": 3,
            "latitudeOfReferencePoint": 49000000,
            "longitudeOfReferencePoint": -2000000,
            "resolutionAndComponentFlags": 0,
            "scaleFactorAtReferencePoint": 0.9996012717,
            "XR": 40000000,
            "YR": -10000000,
            "scanningMode": 64,
            "Di": 200000,
            "Dj": 100000,
            "X1": 29300000,
            "Y1": 9200000,
            "X2": 29900000,
            "Y2": 9400000,
        }
        return section

    def expected(self, y_dim, x_dim, x_negative=False, y_negative=False):
        # Prepare the expectation.
        expected = empty_metadata()
        ellipsoid = iris.coord_systems.GeogCS(6377563.396, 6356256.909)
        cs = iris.coord_systems.TransverseMercator(
            49, -2, 400000, -100000, 0.9996012717, ellipsoid
        )
        nx = 4
        x_origin = 293000
        dx = 2000
        x_array = np.arange(nx) * dx + x_origin
        if x_negative:
            x_array = np.flip(x_array)
        x = iris.coords.DimCoord(
            x_array, "projection_x_coordinate", units="m", coord_system=cs
        )
        ny = 3
        y_origin = 92000
        dy = 1000
        y_array = np.arange(ny) * dy + y_origin
        if y_negative:
            y_array = np.flip(y_array)
        y = iris.coords.DimCoord(
            y_array, "projection_y_coordinate", units="m", coord_system=cs
        )
        expected["dim_coords_and_dims"].append((y, y_dim))
        expected["dim_coords_and_dims"].append((x, x_dim))
        return expected

    def test(self):
        section = self.section_3()
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)

    def test_spherical(self):
        section = self.section_3()
        section["shapeOfTheEarth"] = 0
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        cs = expected["dim_coords_and_dims"][0][0].coord_system
        cs.ellipsoid = iris.coord_systems.GeogCS(6367470)
        self.assertEqual(metadata, expected)

    def test_negative_x(self):
        section = self.section_3()
        section["scanningMode"] = 0b11000000
        section["X1"], section["X2"] = section["X2"], section["X1"]
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1, x_negative=True)
        self.assertEqual(metadata, expected)

    def test_x_inconsistent_direction(self):
        section = self.section_3()
        section["scanningMode"] = 0b11000000
        metadata = empty_metadata()
        with warnings.catch_warnings(record=True) as warn:
            grid_definition_template_12(section, metadata)
        self.assertEqual(len(warn), 1)
        message = "X definition inconsistent: scanningMode"
        self.assertRegex(str(warn[0].message), message)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)

    def test_x_inconsistent_steps(self):
        section = self.section_3()
        section["Ni"] += 1
        metadata = empty_metadata()
        expected_regex = "X definition inconsistent: .* incompatible with step-size"
        with self.assertRaisesRegex(iris.exceptions.TranslationError, expected_regex):
            grid_definition_template_12(section, metadata)

    def test_negative_y(self):
        section = self.section_3()
        section["scanningMode"] = 0b00000000
        section["Y1"], section["Y2"] = section["Y2"], section["Y1"]
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1, y_negative=True)
        self.assertEqual(metadata, expected)

    def test_y_inconsistent_direction(self):
        section = self.section_3()
        section["scanningMode"] = 0b00000000
        metadata = empty_metadata()
        with warnings.catch_warnings(record=True) as warn:
            grid_definition_template_12(section, metadata)
        self.assertEqual(len(warn), 1)
        message = "Y definition inconsistent: scanningMode"
        self.assertRegex(str(warn[0].message), message)
        expected = self.expected(0, 1)
        self.assertEqual(metadata, expected)

    def test_y_inconsistent_steps(self):
        section = self.section_3()
        section["Nj"] += 1
        metadata = empty_metadata()
        expected_regex = "Y definition inconsistent: .* incompatible with step-size"
        with self.assertRaisesRegex(iris.exceptions.TranslationError, expected_regex):
            grid_definition_template_12(section, metadata)

    def test_transposed(self):
        section = self.section_3()
        section["scanningMode"] = 0b01100000
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(1, 0)
        self.assertEqual(metadata, expected)

    def test_di_tolerance(self):
        # Even though Ni * Di doesn't exactly match X1 to X2 it should
        # be close enough to allow the translation.
        section = self.section_3()
        section["X2"] += 1
        metadata = empty_metadata()
        grid_definition_template_12(section, metadata)
        expected = self.expected(0, 1)
        x = expected["dim_coords_and_dims"][1][0]
        x.points = np.linspace(293000, 299000.01, 4)
        self.assertEqual(metadata, expected)

    def test_incompatible_grid_extent(self):
        section = self.section_3()
        section["X2"] += 100
        metadata = empty_metadata()
        with self.assertRaisesRegex(iris.exceptions.TranslationError, "grid"):
            grid_definition_template_12(section, metadata)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function
:func:`iris_grib._load_convert.product_definition_template_15`.

This basically copies code from 'test_product_definition_template_0', and adds
testing for the statistical method and spatial-processing type.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import iris.coords
from iris.exceptions import TranslationError
from iris.coords import CellMethod, DimCoord

from iris_grib.tests.unit.load_convert import empty_metadata
from iris_grib.tests.unit.load_convert import LoadConvertTest
from iris_grib._load_convert import _MDI as MDI

from iris_grib._load_convert import product_definition_template_15


def section_4_sample():
    # Create a dictionary representing a sample section 4 from a grib file.
    return {
        "productDefinitionTemplateNumber": 15,
        "hoursAfterDataCutoff": MDI,
        "minutesAfterDataCutoff": MDI,
        "indicatorOfUnitForForecastTime": 0,  # minutes
        "forecastTime": 360,
        "NV": 0,
        "typeOfFirstFixedSurface": 103,
        "scaleFactorOfFirstFixedSurface": 0,
        "scaledValueOfFirstFixedSurface": 9999,
        "typeOfSecondFixedSurface": 255,
        "statisticalProcess": 2,  # method = maximum
        "spatialProcessing": 0,  # from source grid, no interpolation
        "numberOfPointsUsed": 0,  # no points used because no interpolation
    }


class Test(LoadConvertTest):
    def setUp(self):
        self.time_coord = DimCoord(24, "time", units="hours since epoch")
        self.forecast_period_coord = DimCoord(6, "forecast_period", units="hours")
        self.forecast_ref_time_coord = DimCoord(
            18, "forecast_reference_time", units="hours since epoch"
        )
        self.height_coord = iris.coords.DimCoord(9999, long_name="height", units="m")

    def _translate(self, section):
        # Use pdt 4.15 to populate a metadata dict from the section 4 keys
        metadata = empty_metadata()
        product_definition_template_15(section, metadata, self.time_coord)
        return metadata

    def test_translation(self):
        # Generate metadata from running our sample section through pdt 4.15.
        metadata = self._translate(section_4_sample())

        # Generate a fresh metadata dict and manually populate it with metadata
        # that we expect will be generated from our sample section.
        expected = empty_metadata()
        aux = expected["aux_coords_and_dims"]
        aux.append((self.forecast_period_coord, None))
        aux.append((self.forecast_ref_time_coord, None))
        aux.append((self.time_coord, None))
        aux.append((self.height_coord, None))

        expected["cell_methods"] = [CellMethod(coords=("area",), method="maximum")]
        expected["attributes"]["spatial_processing_type"] = "No interpolation"

        # Now check that the section conversion produces the metadata we
        # expect.
        self.assertMetadataEqual(metadata, expected)

    def test_bad_statistic_method(self):
        section = section_4_sample()
        section["statisticalProcess"] = 999
        msg = r"unsupported statistical process type \[999\]"
        with self.assertRaisesRegex(TranslationError, msg):
            self._translate(section)

    def test_bad_spatial_processing_code(self):
        section = section_4_sample()
        section["spatialProcessing"] = 999
        msg = r"unsupported spatial processing type \[999\]"
        with self.assertRaisesRegex(TranslationError, msg):
            self._translate(section)


if __name__ == "__main__":
    tests.main()
